@@ -37,20 +37,49 @@
 #include <gnunet/gnunet_common.h>
 #include <gnunet/gnunet_time_lib.h>
 #include <gnunet/gnunet_util_lib.h>
+#include <stdint.h>
 
 /**
  * @def GNUNET_CHAT_VERSION The major and minor version should be identical to
  *                          the #GNUNET_MESSENGER_VERSION of the GNUnet Messenger
  *                          service while the patch version is independent.
  */
-#define GNUNET_CHAT_VERSION 0x000000030001L
+#define GNUNET_CHAT_VERSION 0x000000050000L
 
 #define GNUNET_CHAT_VERSION_MAJOR ((GNUNET_CHAT_VERSION >> 32L) & 0xFFFFL)
 #define GNUNET_CHAT_VERSION_MINOR ((GNUNET_CHAT_VERSION >> 16L) & 0xFFFFL)
 #define GNUNET_CHAT_VERSION_PATCH ((GNUNET_CHAT_VERSION >>  0L) & 0xFFFFL)
 
+#define GNUNET_CHAT_ATTRIBUTE_AVATAR   "chat_avatar"
+#define GNUNET_CHAT_ATTRIBUTE_BIRTHDAY "chat_birthday"
+#define GNUNET_CHAT_ATTRIBUTE_EMAIL    "chat_email"
+#define GNUNET_CHAT_ATTRIBUTE_NAME     "chat_name"
+#define GNUNET_CHAT_ATTRIBUTE_PHONE    "chat_phone"
+#define GNUNET_CHAT_ATTRIBUTE_POSTAL   "chat_postal"
+
 #define GNUNET_CHAT_URI_PREFIX "gnunet://chat/"
 
+/**
+ * Enum for the different types of supported URIs.
+ */
+enum GNUNET_CHAT_UriType
+{
+  /**
+   * The type to share and join a lobby or chat.
+   */
+  GNUNET_CHAT_URI_TYPE_CHAT = 1,   /**< GNUNET_CHAT_URI_TYPE_CHAT */
+
+  /**
+   * The type to download and query files vis FS service from GNUnet.
+   */
+  GNUNET_CHAT_URI_TYPE_FS = 2,     /**< GNUNET_CHAT_URI_TYPE_FS */
+
+  /**
+   * An unknown type of uri.
+   */
+  GNUNET_CHAT_URI_TYPE_UNKNOWN = 0 /**< GNUNET_CHAT_URI_TYPE_UNKNOWN */
+};
+
 /**
  * Enum for the different kinds of messages.
  */
@@ -59,73 +88,108 @@ enum GNUNET_CHAT_MessageKind
   /**
    * The kind to inform that something went wrong.
    */
-  GNUNET_CHAT_KIND_WARNING = 1,    /**< GNUNET_CHAT_KIND_WARNING */
+  GNUNET_CHAT_KIND_WARNING = 1,            /**< GNUNET_CHAT_KIND_WARNING */
 
   /**
    * The kind to inform that the list of accounts was refreshed.
    */
-  GNUNET_CHAT_KIND_REFRESH = 2,    /**< GNUNET_CHAT_KIND_REFRESH */
+  GNUNET_CHAT_KIND_REFRESH = 2,            /**< GNUNET_CHAT_KIND_REFRESH */
 
   /**
    * The kind to inform that the application can be used.
    */
-  GNUNET_CHAT_KIND_LOGIN = 3,      /**< GNUNET_CHAT_KIND_LOGIN */
+  GNUNET_CHAT_KIND_LOGIN = 3,              /**< GNUNET_CHAT_KIND_LOGIN */
 
   /**
    * The kind to inform that the application needs to cleanup 
    * resources related to the currently connected account.
    */
-  GNUNET_CHAT_KIND_LOGOUT = 4,     /**< GNUNET_CHAT_KIND_LOGOUT */
+  GNUNET_CHAT_KIND_LOGOUT = 4,             /**< GNUNET_CHAT_KIND_LOGOUT */
+
+  /**
+   * The kind to inform that an account was successfully created.
+   */
+  GNUNET_CHAT_KIND_CREATED_ACCOUNT = 5,    /**< GNUNET_CHAT_KIND_CREATED_ACCOUNT */
+
+  /**
+   * The kind to inform that an account was successfully deleted.
+   */
+  GNUNET_CHAT_KIND_DELETED_ACCOUNT = 6,    /**< GNUNET_CHAT_KIND_DELETED_ACCOUNT */
+
+  /**
+   * The kind to inform that an account was updated.
+   */
+  GNUNET_CHAT_KIND_UPDATE_ACCOUNT = 7,     /**< GNUNET_CHAT_KIND_UPDATE_ACCOUNT */
 
   /**
    * The kind to inform that a context was updated.
    */
-  GNUNET_CHAT_KIND_UPDATE = 5,     /**< GNUNET_CHAT_KIND_UPDATE */
+  GNUNET_CHAT_KIND_UPDATE_CONTEXT = 8,     /**< GNUNET_CHAT_KIND_UPDATE_CONTEXT */
 
   /**
    * The kind to inform that a contact has joined a chat.
    */
-  GNUNET_CHAT_KIND_JOIN = 6,       /**< GNUNET_CHAT_KIND_JOIN */
+  GNUNET_CHAT_KIND_JOIN = 9,               /**< GNUNET_CHAT_KIND_JOIN */
 
   /**
    * The kind to inform that a contact has left a chat.
    */
-  GNUNET_CHAT_KIND_LEAVE = 7,      /**< GNUNET_CHAT_KIND_LEAVE */
+  GNUNET_CHAT_KIND_LEAVE = 10,             /**< GNUNET_CHAT_KIND_LEAVE */
 
   /**
    * The kind to inform that a contact has changed.
    */
-  GNUNET_CHAT_KIND_CONTACT = 8,    /**< GNUNET_CHAT_KIND_CONTACT */
+  GNUNET_CHAT_KIND_CONTACT = 11,           /**< GNUNET_CHAT_KIND_CONTACT */
 
   /**
    * The kind to describe an invitation to a different chat.
    */
-  GNUNET_CHAT_KIND_INVITATION = 9, /**< GNUNET_CHAT_KIND_INVITATION */
+  GNUNET_CHAT_KIND_INVITATION = 12,        /**< GNUNET_CHAT_KIND_INVITATION */
 
   /**
    * The kind to describe a text message.
    */
-  GNUNET_CHAT_KIND_TEXT = 10,      /**< GNUNET_CHAT_KIND_TEXT */
+  GNUNET_CHAT_KIND_TEXT = 13,              /**< GNUNET_CHAT_KIND_TEXT */
 
   /**
    * The kind to describe a shared file.
    */
-  GNUNET_CHAT_KIND_FILE = 11,      /**< GNUNET_CHAT_KIND_FILE */
+  GNUNET_CHAT_KIND_FILE = 14,              /**< GNUNET_CHAT_KIND_FILE */
 
   /**
    * The kind to inform about a deletion of a previous message.
    */
-  GNUNET_CHAT_KIND_DELETION = 12,  /**< GNUNET_CHAT_KIND_DELETION */
+  GNUNET_CHAT_KIND_DELETION = 15,          /**< GNUNET_CHAT_KIND_DELETION */
 
   /**
    * The kind to tag a previous message.
    */
-  GNUNET_CHAT_KIND_TAG = 13,       /**< GNUNET_CHAT_KIND_TAG */
+  GNUNET_CHAT_KIND_TAG = 16,               /**< GNUNET_CHAT_KIND_TAG */
+
+  /**
+   * The kind to inform that attributes were updated.
+   */
+  GNUNET_CHAT_KIND_ATTRIBUTES = 17,        /**< GNUNET_CHAT_KIND_ATTRIBUTES */
+
+  /**
+   * The kind to inform that attributes were shared.
+   */
+  GNUNET_CHAT_KIND_SHARED_ATTRIBUTES = 18, /**< GNUNET_CHAT_KIND_SHARED_ATTRIBUTES */
+
+  /**
+   * The kind to inform that a discourse was updated.
+   */
+  GNUNET_CHAT_KIND_DISCOURSE = 19,         /**< GNUNET_CHAT_KIND_DISCOURSE */
+
+  /**
+   * The kind to describe a data message from a discourse.
+   */
+  GNUNET_CHAT_KIND_DATA = 20,              /**< GNUNET_CHAT_KIND_DATA */
 
   /**
    * An unknown kind of message.
    */
-  GNUNET_CHAT_KIND_UNKNOWN = 0     /**< GNUNET_CHAT_KIND_UNKNOWN */
+  GNUNET_CHAT_KIND_UNKNOWN = 0             /**< GNUNET_CHAT_KIND_UNKNOWN */
 };
 
 /**
@@ -179,9 +243,9 @@ struct GNUNET_CHAT_File;
 struct GNUNET_CHAT_Invitation;
 
 /**
- * Struct of a chat ticket.
+ * Struct of a chat discourse.
  */
-struct GNUNET_CHAT_Ticket;
+struct GNUNET_CHAT_Discourse;
 
 /**
  * Iterator over chat accounts of a specific chat handle.
@@ -197,14 +261,30 @@ typedef enum GNUNET_GenericReturnValue
                                 struct GNUNET_CHAT_Account *account);
 
 /**
- * Method called for each attribute of a specific chat handle.
+ * Iterator over attributes of a specific chat account.
  *
  * @param[in,out] cls Closure from #GNUNET_CHAT_get_attributes
- * @param[in] handle Chat handle
+ * @param[in] account Chat account
  * @param[in] name Attribute name
  * @param[in] value Attribute value
+ * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
-typedef void
+typedef enum GNUNET_GenericReturnValue
+(*GNUNET_CHAT_AccountAttributeCallback) (void *cls,
+                                         const struct GNUNET_CHAT_Account *account,
+                                         const char *name,
+                                         const char *value);
+
+/**
+ * Iterator over attributes of a specific chat handle.
+ *
+ * @param[in,out] cls Closure from #GNUNET_CHAT_get_attributes
+ * @param[in,out] handle Chat handle
+ * @param[in] name Attribute name
+ * @param[in] value Attribute value
+ * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
+ */
+typedef enum GNUNET_GenericReturnValue
 (*GNUNET_CHAT_AttributeCallback) (void *cls,
                                   struct GNUNET_CHAT_Handle *handle,
                                   const char *name,
@@ -220,6 +300,19 @@ typedef void
 (*GNUNET_CHAT_LobbyCallback) (void *cls, 
                               const struct GNUNET_CHAT_Uri *uri);
 
+/**
+ * Iterator over chat files of a specific chat handle.
+ *
+ * @param[in,out] cls Closure from #GNUNET_CHAT_iterate_files
+ * @param[in,out] handle Chat handle
+ * @param[in,out] file Chat file
+ * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
+ */
+typedef enum GNUNET_GenericReturnValue
+(*GNUNET_CHAT_FileCallback) (void *cls,
+                             struct GNUNET_CHAT_Handle *handle,
+                             struct GNUNET_CHAT_File *file);
+
 /**
  * Iterator over chat contacts of a specific chat handle.
  *
@@ -234,17 +327,19 @@ typedef enum GNUNET_GenericReturnValue
                                 struct GNUNET_CHAT_Contact *contact);
 
 /**
- * Iterator over chat tickets of a specific chat contact.
+ * Iterator over accessible attributes of a specific chat contact.
  *
  * @param[in,out] cls Closure
- * @param[in] contact Chat contact
- * @param[in,out] ticket Chat ticket
+ * @param[in,out] contact Chat contact
+ * @param[in] name Attribute name
+ * @param[in] value Attribute value
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
 typedef enum GNUNET_GenericReturnValue
-(*GNUNET_CHAT_ContactTicketCallback) (void *cls,
-                                      const struct GNUNET_CHAT_Contact *contact,
-                                      struct GNUNET_CHAT_Ticket *ticket);
+(*GNUNET_CHAT_ContactAttributeCallback) (void *cls,
+                                         struct GNUNET_CHAT_Contact *contact,
+                                         const char *name,
+                                         const char *value);
 
 /**
  * Iterator over chat groups of a specific chat handle.
@@ -263,7 +358,7 @@ typedef enum GNUNET_GenericReturnValue
  * Iterator over chat contacts in a specific chat group.
  *
  * @param[in,out] cls Closure from #GNUNET_CHAT_group_iterate_contacts
- * @param[in,out] group Chat group
+ * @param[in] group Chat group
  * @param[in,out] contact Chat contact
  * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
  */
@@ -330,13 +425,13 @@ typedef enum GNUNET_GenericReturnValue
  * Method called during an upload of a specific file in a chat to share it.
  *
  * @param[in,out] cls Closure from #GNUNET_CHAT_context_send_file
- * @param[in] file Chat file
+ * @param[in,out] file Chat file
  * @param[in] completed Amount of the file being uploaded (in bytes)
  * @param[in] size Full size of the uploading file (in bytes)
  */
 typedef void
 (*GNUNET_CHAT_FileUploadCallback) (void *cls,
-                                   const struct GNUNET_CHAT_File *file,
+                                   struct GNUNET_CHAT_File *file,
                                    uint64_t completed,
                                    uint64_t size);
 
@@ -344,13 +439,13 @@ typedef void
  * Method called during a download of a specific file in a chat which was shared.
  *
  * @param[in,out] cls Closure from #GNUNET_CHAT_file_start_download
- * @param[in] file Chat file
+ * @param[in,out] file Chat file
  * @param[in] completed Amount of the file being downloaded (in bytes)
  * @param[in] size Full size of the downloading file (in bytes)
  */
 typedef void
 (*GNUNET_CHAT_FileDownloadCallback) (void *cls,
-                                     const struct GNUNET_CHAT_File *file,
+                                     struct GNUNET_CHAT_File *file,
                                      uint64_t completed,
                                      uint64_t size);
 
@@ -370,18 +465,29 @@ typedef void
                                     uint64_t size);
 
 /**
- * Method called during a chat ticket consumption for each of its attributes.
+ * Iterator over chat discourses in a specific chat context.
  *
- * @param[in,out] cls Closure from #GNUNET_CHAT_ticket_consume
- * @param[in] ticket Chat ticket
- * @param[in] name Attribute name
- * @param[in] value Attribute value
+ * @param[in,out] cls Closure from #GNUNET_CHAT_context_iterate_discourses
+ * @param[in,out] context Chat context
+ * @param[in,out] discourse Chat discourse
  */
-typedef void
-(*GNUNET_CHAT_TicketAttributeCallback) (void *cls,
-                                        const struct GNUNET_CHAT_Ticket *ticket,
-                                        const char *name,
-                                        const char *value);
+typedef enum GNUNET_GenericReturnValue
+(*GNUNET_CHAT_DiscourseCallback) (void *cls,
+                                  struct GNUNET_CHAT_Context *context,
+                                  struct GNUNET_CHAT_Discourse *discourse);
+
+/**
+ * Iterator over chat contacts in a specific chat discourse.
+ *
+ * @param[in,out] cls Closure from #GNUNET_CHAT_discourse_iterate_contacts
+ * @param[in] discourse Chat discourse
+ * @param[in,out] contact Chat contact
+ * @return #GNUNET_YES if we should continue to iterate, #GNUNET_NO otherwise.
+ */
+typedef enum GNUNET_GenericReturnValue
+(*GNUNET_CHAT_DiscourseContactCallback) (void *cls,
+                                         const struct GNUNET_CHAT_Discourse *discourse,
+                                         struct GNUNET_CHAT_Contact *contact);
 
 /**
  * Start a chat handle with a certain configuration.
@@ -532,11 +638,11 @@ void
 GNUNET_CHAT_set_attribute (struct GNUNET_CHAT_Handle *handle,
                            const char *name,
                            const char *value,
-                           const struct GNUNET_TIME_Relative *expires);
+                           struct GNUNET_TIME_Relative expires);
 
 /**
- * Deletes an attribute of a chat handle for related communication under a given
- * <i>name</i>.
+ * Deletes an attribute of a chat <i>handle</i> for related communication 
+ * under a given <i>name</i>.
  *
  * @param[in,out] handle Chat handle
  * @param[in] name Attribute name
@@ -558,12 +664,53 @@ GNUNET_CHAT_get_attributes (struct GNUNET_CHAT_Handle *handle,
                             GNUNET_CHAT_AttributeCallback callback,
                             void *cls);
 
+/**
+ * Share an attribute of a chat <i>handle</i> under a given <i>name</i>
+ * with a specific chat <i>contact</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in,out] contact Chat contact
+ * @param[in] name Attribute name
+ */
+void
+GNUNET_CHAT_share_attribute_with (struct GNUNET_CHAT_Handle *handle,
+                                  struct GNUNET_CHAT_Contact *contact,
+                                  const char *name);
+
+/**
+ * Unshare an attribute of a chat <i>handle</i> under a given <i>name</i>
+ * from a specific chat <i>contact</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in,out] contact Chat contact
+ * @param[in] name Attribute name
+ */
+void
+GNUNET_CHAT_unshare_attribute_from (struct GNUNET_CHAT_Handle *handle,
+                                    struct GNUNET_CHAT_Contact *contact,
+                                    const char *name);
+
+/**
+ * Calls an optional <i>callback</i> for each attribute of a given chat 
+ * <i>handle</i> shared with a specific chat <i>contact</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] contact Chat contact
+ * @param[in] callback Callback for attribute iteration (optional)
+ * @param[in,out] cls Closure for attribute iteration (optional)
+ */
+void
+GNUNET_CHAT_get_shared_attributes (struct GNUNET_CHAT_Handle *handle,
+                                   struct GNUNET_CHAT_Contact *contact,
+                                   GNUNET_CHAT_ContactAttributeCallback callback,
+                                   void *cls);
+
 /**
  * Convert an UTF-8 String to a chat URI which will be newly allocated.
  *
  * @param[in] uri UTF-8 string to parse
  * @param[out] emsg Where to store the parser error message (if any)
- * @return NULL on error
+ * @return URI on success, NULL on error
  */
 struct GNUNET_CHAT_Uri*
 GNUNET_CHAT_uri_parse (const char *uri,
@@ -578,6 +725,15 @@ GNUNET_CHAT_uri_parse (const char *uri,
 char*
 GNUNET_CHAT_uri_to_string (const struct GNUNET_CHAT_Uri *uri);
 
+/**
+ * Returns the type of a given chat URI.
+ *
+ * @param[in] uri Chat URI
+ * @return The type of URI
+ */
+enum GNUNET_CHAT_UriType
+GNUNET_CHAT_uri_get_type (const struct GNUNET_CHAT_Uri *uri);
+
 /**
  * Free an allocated chat URI.
  *
@@ -620,6 +776,47 @@ void
 GNUNET_CHAT_lobby_join (struct GNUNET_CHAT_Handle *handle,
                         const struct GNUNET_CHAT_Uri *uri);
 
+/**
+ * Requests a file with a given chat <i>handle</i> from a selected chat 
+ * <i>uri</i> to potentially download it.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] uri Chat URI
+ * @return File handle on success, NULL on failure
+ */
+struct GNUNET_CHAT_File*
+GNUNET_CHAT_request_file (struct GNUNET_CHAT_Handle *handle,
+                          const struct GNUNET_CHAT_Uri *uri);
+
+/**
+ * Uploads a local file specified via its <i>path</i> to share its uri
+ * with others afterwards.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] path Local file path
+ * @param[in] callback Callback for file uploading (optional)
+ * @param[in,out] cls Closure for file uploading (optional)
+ * @return The file handle on success, NULL on failure
+ */
+struct GNUNET_CHAT_File*
+GNUNET_CHAT_upload_file (struct GNUNET_CHAT_Handle *handle,
+                         const char *path,
+                         GNUNET_CHAT_FileUploadCallback callback,
+                         void *cls);
+
+/**
+ * Iterates through the files of a given chat <i>handle</i> with a selected
+ * callback and custom closure.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] callback Callback for file iteration (optional)
+ * @param[in,out] cls Closure for file iteration (optional)
+ * @return Amount of files iterated or #GNUNET_SYSERR on failure
+ */
+int
+GNUNET_CHAT_iterate_files (struct GNUNET_CHAT_Handle *handle,
+                           GNUNET_CHAT_FileCallback callback,
+                           void *cls);
 
 /**
  * Sets a custom <i>user pointer</i> to a given chat <i>handle</i> so it can
@@ -665,6 +862,21 @@ GNUNET_CHAT_iterate_contacts (struct GNUNET_CHAT_Handle *handle,
 const char*
 GNUNET_CHAT_account_get_name (const struct GNUNET_CHAT_Account *account);
 
+/**
+ * Calls an optional <i>callback</i> for each attribute of a given chat 
+ * <i>account</i> using a chat <i>handle</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] account Chat account
+ * @param[in] callback Callback for attribute iteration (optional)
+ * @param[in,out] cls Closure for attribute iteration (optional)
+ */
+void
+GNUNET_CHAT_account_get_attributes (struct GNUNET_CHAT_Handle *handle,
+                                    const struct GNUNET_CHAT_Account *account,
+                                    GNUNET_CHAT_AccountAttributeCallback callback,
+                                    void *cls);
+
 /**
  * Sets a custom <i>user pointer</i> to a given chat <i>account</i> so it can
  * be accessed in account related callbacks.
@@ -855,18 +1067,17 @@ GNUNET_CHAT_contact_is_tagged (const struct GNUNET_CHAT_Contact *contact,
                                const char *tag);
 
 /**
- * Iterates through the tickets of a given <i>contact</i> with a selected
- * callback and custom closure.
+ * Calls an optional <i>callback</i> for each attribute of a given chat 
+ * <i>contact</i>.
  *
- * @param[in,out] contact Contact
- * @param[in] callback Callback for ticket iteration (optional)
- * @param[in,out] cls Closure for ticket iteration (optional)
- * @return Amount of tickets iterated
+ * @param[in,out] contact Chat contact
+ * @param[in] callback Callback for attribute iteration (optional)
+ * @param[in,out] cls Closure for attribute iteration (optional)
  */
-int
-GNUNET_CHAT_contact_iterate_tickets (const struct GNUNET_CHAT_Contact *contact,
-                                     GNUNET_CHAT_ContactTicketCallback callback,
-                                     void *cls);
+void
+GNUNET_CHAT_contact_get_attributes (struct GNUNET_CHAT_Contact *contact,
+                                    GNUNET_CHAT_ContactAttributeCallback callback,
+                                    void *cls);
 
 /**
  * Leaves a specific <i>group</i> chat and frees its memory if it is not shared
@@ -1105,6 +1316,18 @@ GNUNET_CHAT_context_send_tag (struct GNUNET_CHAT_Context *context,
                               const struct GNUNET_CHAT_Message *message,
                               const char *tag);
 
+/**
+ * Opens a chat discourse under a specific <i>id</i> in a given chat 
+ * <i>context</i> to send data live to other contacts.
+ *
+ * @param[in,out] context Chat context
+ * @param[in] id Discourse id
+ * @return Chat discourse
+ */
+struct GNUNET_CHAT_Discourse*
+GNUNET_CHAT_context_open_discourse (struct GNUNET_CHAT_Context *context,
+                                    const struct GNUNET_ShortHashCode *id);
+
 /**
  * Iterates through the contacts of a given chat <i>context</i> with a selected
  * callback and custom closure.
@@ -1133,6 +1356,20 @@ GNUNET_CHAT_context_iterate_files (struct GNUNET_CHAT_Context *context,
                                    GNUNET_CHAT_ContextFileCallback callback,
                                    void *cls);
 
+/**
+ * Iterates through the discourses of a given chat <i>context</i> with a 
+ * selected callback and custom closure.
+ *
+ * @param[in,out] context Chat context
+ * @param[in] callback Callback for file iteration (optional)
+ * @param[in,out] cls Closure for file iteration (optional)
+ * @return Amount of discourses iterated or #GNUNET_SYSERR on failure
+ */
+int
+GNUNET_CHAT_context_iterate_discourses (struct GNUNET_CHAT_Context *context,
+                                        GNUNET_CHAT_DiscourseCallback callback,
+                                        void *cls);
+
 /**
  * Returns the kind of a given <i>message</i> to determine its content and
  * related usage.
@@ -1262,6 +1499,16 @@ GNUNET_CHAT_message_get_read_receipt (const struct GNUNET_CHAT_Message *message,
 const char*
 GNUNET_CHAT_message_get_text (const struct GNUNET_CHAT_Message *message);
 
+/**
+ * Returns the account of a given <i>message</i> which is either 
+ * its sender or target of the message depending on the kind of
+ * the message, otherwise it returns NULL.
+ *
+ * @param[in] message Message
+ * @return The account of message or NULL
+ */
+const struct GNUNET_CHAT_Account*
+GNUNET_CHAT_message_get_account (const struct GNUNET_CHAT_Message *message);
 
 /**
  * Returns the file handle of a given <i>message</i> if its kind is
@@ -1283,6 +1530,17 @@ GNUNET_CHAT_message_get_file (const struct GNUNET_CHAT_Message *message);
 struct GNUNET_CHAT_Invitation*
 GNUNET_CHAT_message_get_invitation (const struct GNUNET_CHAT_Message *message);
 
+/**
+ * Returns the discourse of a given <i>message</i> if its kind is
+ * #GNUNET_CHAT_KIND_DISCOURSE or #GNUNET_CHAT_KIND_DATA,
+ * otherwise it returns NULL.
+ *
+ * @param[in] message Message
+ * @return The discourse of message or NULL
+ */
+struct GNUNET_CHAT_Discourse*
+GNUNET_CHAT_message_get_discourse (const struct GNUNET_CHAT_Message *message);
+
 /**
  * Returns the target message of an operation represented by a given
  * <i>message</i> if its kind is #GNUNET_CHAT_KIND_DELETION, otherwise it
@@ -1319,6 +1577,32 @@ GNUNET_CHAT_message_iterate_tags (const struct GNUNET_CHAT_Message *message,
                                   GNUNET_CHAT_MessageCallback callback,
                                   void *cls);
 
+/**
+ * Returns the amount of data from a given <i>message</i> if its kind 
+ * is #GNUNET_CHAT_KIND_DATA, otherwise it returns NULL.
+ *
+ * @param[in] message Message
+ * @return The amount of data or zero
+ */
+uint64_t
+GNUNET_CHAT_message_available (const struct GNUNET_CHAT_Message *message);
+
+/**
+ * Reads the data from a given <i>message</i> if its kind is
+ * #GNUNET_CHAT_KIND_DATA into a specific <i>data</i> buffer
+ * up to a selected <i>size</i>.
+ *
+ * @param[in] message Message
+ * @param[out] data Data buffer
+ * @param[in] size Data size
+ * @return #GNUNET_OK on success, #GNUNET_NO if there's missing data 
+ *         to read, otherwise #GNUNET_SYSERR on failure
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_message_read (const struct GNUNET_CHAT_Message *message,
+                          char *data,
+                          uint64_t size);
+
 /**
  * Returns the name of a given <i>file</i> handle.
  *
@@ -1358,6 +1642,16 @@ GNUNET_CHAT_file_get_size (const struct GNUNET_CHAT_File *file);
 uint64_t
 GNUNET_CHAT_file_get_local_size (const struct GNUNET_CHAT_File *file);
 
+/**
+ * Returns a new allocated uri to access the content of a given 
+ * <i>file</i> handle.
+ *
+ * @param[in] file File handle
+ * @return URI on success, NULL on error
+ */
+struct GNUNET_CHAT_Uri*
+GNUNET_CHAT_file_get_uri (const struct GNUNET_CHAT_File *file);
+
 /**
  * Returns if a given <i>file</i> handle is currently uploading.
  *
@@ -1367,6 +1661,16 @@ GNUNET_CHAT_file_get_local_size (const struct GNUNET_CHAT_File *file);
 enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_is_uploading (const struct GNUNET_CHAT_File *file);
 
+/**
+ * Returns if a given <i>file</i> handle is currently ready
+ * for previewing its content.
+ *
+ * @param[in] file File handle
+ * @return #GNUNET_YES when ready, #GNUNET_NO otherwise
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_file_is_ready (const struct GNUNET_CHAT_File *file);
+
 /**
  * Returns the temporary file name of the decrypted file preview
  * of a given <i>file</i> handle.
@@ -1521,26 +1825,60 @@ enum GNUNET_GenericReturnValue
 GNUNET_CHAT_invitation_is_rejected (const struct GNUNET_CHAT_Invitation *invitation);
 
 /**
- * Returns the contact of the issuer from a given chat <i>ticket</i>.
+ * Returns the discourse id of a given chat <i>discourse</i>.
  *
- * @param[in] invitation Chat invitation
- * @return Chat contact
+ * @param[in] discourse Chat discourse
+ * @return Discourse id
  */
-const struct GNUNET_CHAT_Contact*
-GNUNET_CHAT_ticket_get_contact (const struct GNUNET_CHAT_Ticket *ticket);
+const struct GNUNET_ShortHashCode*
+GNUNET_CHAT_discourse_get_id (const struct GNUNET_CHAT_Discourse *discourse);
 
 /**
- * Consumes a given chat <i>ticket</i> and calls an optional <i>callback</i>
- * for each of its attributes.
+ * Returns whether a chat <i>discourse</i> is currently open.
  *
- * @param[in,out] ticket Chat ticket
- * @param[in] callback Callback for ticket consumption (optional)
- * @param[in,out] cls Closure for ticket consumption (optional)
+ * @param[in] discourse Chat discourse
+ * @return #GNUNET_YES if the discourse is open, #GNUNET_SYSERR on failure and 
+ *         #GNUNET_NO otherwise.
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_discourse_is_open (const struct GNUNET_CHAT_Discourse *discourse);
+
+/**
+ * Closes a given chat <i>discourse</i> to stop receiving any data messages
+ * from the specific discourse.
+ *
+ * @param[in,out] discourse Chat discourse
  */
 void
-GNUNET_CHAT_ticket_consume (struct GNUNET_CHAT_Ticket *ticket,
-                            GNUNET_CHAT_TicketAttributeCallback callback,
-                            void *cls);
+GNUNET_CHAT_discourse_close (struct GNUNET_CHAT_Discourse *discourse);
+
+/**
+ * Sends messages to a given chat <i>discourse</i> containing the
+ * specified <i>data</i> of a custom <i>size</i> in bytes.
+ *
+ * @param[in,out] discourse Chat discourse
+ * @param[in] data Data buffer
+ * @param[in] size Data size
+ * @return #GNUNET_OK on success, otherwise #GNUNET_SYSERR
+ */
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_discourse_write (struct GNUNET_CHAT_Discourse *discourse,
+                             const char *data,
+                             uint64_t size);
+
+/**
+ * Iterates through the subscribed chat contacts of a given chat <i>discourse</i> 
+ * with a selected callback and custom closure.
+ *
+ * @param[in,out] discourse Chat discourse
+ * @param[in] callback Callback for contact iteration (optional)
+ * @param[in,out] cls Closure for contact iteration (optional)
+ * @return Amount of contacts iterated or #GNUNET_SYSERR on failure
+ */
+int
+GNUNET_CHAT_discourse_iterate_contacts (const struct GNUNET_CHAT_Discourse *discourse,
+                                        GNUNET_CHAT_DiscourseContactCallback callback,
+                                        void *cls);
 
 /**@}*/
 
