@@ -25,9 +25,12 @@
 #ifndef GNUNET_CHAT_ACCOUNT_H_
 #define GNUNET_CHAT_ACCOUNT_H_
 
+#include <gnunet/gnunet_common.h>
 #include <gnunet/gnunet_identity_service.h>
 #include <gnunet/gnunet_util_lib.h>
 
+struct GNUNET_CHAT_Handle;
+
 struct GNUNET_CHAT_Account
 {
   struct GNUNET_IDENTITY_Ego *ego;
@@ -37,6 +40,15 @@ struct GNUNET_CHAT_Account
   void *user_pointer;
 };
 
+/**
+ * Creates a chat account using a given <i>name</i>.
+ *
+ * @param[in] name Name
+ * @return New chat account
+ */
+struct GNUNET_CHAT_Account*
+account_create (const char *name);
+
 /**
  * Creates a chat account using a given <i>ego</i> and
  * a matching <i>name</i>.
@@ -46,8 +58,8 @@ struct GNUNET_CHAT_Account
  * @return New chat account
  */
 struct GNUNET_CHAT_Account*
-account_create_from_ego(struct GNUNET_IDENTITY_Ego *ego,
-			                  const char *name);
+account_create_from_ego (struct GNUNET_IDENTITY_Ego *ego,
+			                   const char *name);
 
 /**
  * Updates the stored directory path by a chat <i>account</i>
@@ -60,12 +72,45 @@ void
 account_update_directory (struct GNUNET_CHAT_Account *account,
 			                    const char *base_directory);
 
+/**
+ * Returns the private key from a given chat
+ * <i>account</i>.
+ *
+ * @param[in] account Chat account
+ * @return EGOs private key or NULL
+ */
+const struct GNUNET_CRYPTO_PrivateKey*
+account_get_key (const struct GNUNET_CHAT_Account *account);
+
+/**
+ * Updates the key from a given chat <i>account</i> using
+ * the chat <i>handle</i> and a specific <i>ego</i> matching 
+ * the accounts name.
+ *
+ * @param[in,out] account Chat account
+ * @param[in,out] handle Chat handle
+ * @param[in] ego EGO
+ */
+void
+account_update_ego (struct GNUNET_CHAT_Account *account,
+                    struct GNUNET_CHAT_Handle *handle,
+                    struct GNUNET_IDENTITY_Ego *ego);
+
+/**
+ * Deletes all local files and data remaining a given 
+ * chat <i>account</i>.
+ *
+ * @param[in,out] account Chat account
+ */
+void
+account_delete (struct GNUNET_CHAT_Account *account);
+
 /**
  * Destroys a chat <i>account</i> and frees its memory.
  *
  * @param[in,out] account Chat account
  */
 void
-account_destroy(struct GNUNET_CHAT_Account *account);
+account_destroy (struct GNUNET_CHAT_Account *account);
 
 #endif /* GNUNET_CHAT_ACCOUNT_H_ */
