@@ -127,24 +127,34 @@ contact_update_key (struct GNUNET_CHAT_Contact *contact)
   if (contact->public_key)
     GNUNET_free(contact->public_key);
 
-  contact->public_key = NULL;
-
-  if (!(contact->member))
-    return;
-
   const struct GNUNET_CRYPTO_PublicKey *pubkey;
-  pubkey = GNUNET_MESSENGER_contact_get_key(contact->member);
+  pubkey = contact_get_key(contact);
 
   if (pubkey)
     contact->public_key = GNUNET_CRYPTO_public_key_to_string(pubkey);
+  else
+    contact->public_key = NULL;
 }
 
-struct GNUNET_CHAT_Context*
-contact_find_context (const struct GNUNET_CHAT_Contact *contact)
+const struct GNUNET_CRYPTO_PublicKey*
+contact_get_key (const struct GNUNET_CHAT_Contact *contact)
 {
   GNUNET_assert(contact);
 
-  if (contact->context)
+  if (!(contact->member))
+    return NULL;
+
+  return GNUNET_MESSENGER_contact_get_key(contact->member);
+}
+
+struct GNUNET_CHAT_Context*
+contact_find_context (const struct GNUNET_CHAT_Contact *contact,
+                      enum GNUNET_GenericReturnValue room_required)
+{
+  GNUNET_assert(contact);
+
+  if ((contact->context) &&
+      ((GNUNET_YES != room_required) || (contact->context->room)))
     return contact->context;
 
   struct GNUNET_CHAT_ContactFindRoom find;
@@ -161,10 +171,15 @@ contact_find_context (const struct GNUNET_CHAT_Contact *contact)
   if (!(find.room))
     return NULL;
 
-  return GNUNET_CONTAINER_multihashmap_get(
+  struct GNUNET_CHAT_Context *context = GNUNET_CONTAINER_multihashmap_get(
     contact->handle->contexts,
     GNUNET_MESSENGER_room_get_key(find.room)
   );
+
+  if ((GNUNET_YES == room_required) && (!(context->room)))
+    return NULL;
+
+  return context;
 }
 
 const struct GNUNET_HashCode*
@@ -352,6 +367,8 @@ skip_tag_search:
   char *tag_value = tag? GNUNET_strdup(tag) : NULL;
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
   msg.header.kind = GNUNET_MESSENGER_KIND_TAG;
   GNUNET_memcpy(&(msg.body.tag.hash), hash,
     sizeof(struct GNUNET_HashCode));
@@ -400,7 +417,7 @@ contact_destroy (struct GNUNET_CHAT_Contact* contact)
     GNUNET_CONTAINER_multihashmap_destroy(contact->joined);
   }
 
-  if ((contact->context) && (!contact->context->room))
+  if ((contact->context) && (!(contact->context->room)))
     context_destroy(contact->context);
 
   GNUNET_free(contact);
