@@ -29,13 +29,17 @@
 #include <gnunet/gnunet_messenger_service.h>
 #include <gnunet/gnunet_util_lib.h>
 
+#include "gnunet_chat_lib.h"
+
 struct GNUNET_CHAT_Handle;
+struct GNUNET_CHAT_Contact;
 struct GNUNET_CHAT_Context;
 struct GNUNET_CHAT_Ticket;
 
 struct GNUNET_CHAT_InternalTickets
 {
   struct GNUNET_CHAT_Ticket *ticket;
+
   struct GNUNET_CHAT_InternalTickets *next;
   struct GNUNET_CHAT_InternalTickets *prev;
 };
@@ -93,15 +97,25 @@ contact_update_join (struct GNUNET_CHAT_Contact *contact,
 void
 contact_update_key (struct GNUNET_CHAT_Contact *contact);
 
+/**
+ * Returns the public key from a given chat <i>contact</i>.
+ *
+ * @param[in] contact Chat contact
+ */
+const struct GNUNET_CRYPTO_PublicKey*
+contact_get_key (const struct GNUNET_CHAT_Contact *contact);
+
 /**
  * Searches for a chat context containing a given chat
  * <i>contact</i> and the least amount of other members.
  *
  * @param[in] contact Chat contact
+ * @param[in] room_required Flag to suggest a room is required
  * @return Chat context or NULL
  */
 struct GNUNET_CHAT_Context*
-contact_find_context (const struct GNUNET_CHAT_Contact *contact);
+contact_find_context (const struct GNUNET_CHAT_Contact *contact,
+                      enum GNUNET_GenericReturnValue room_required);
 
 /**
  * Returns whether a chat <i>contact</i> is tagged in
