@@ -25,9 +25,11 @@
 #include "gnunet_chat_handle.h"
 
 #include "gnunet_chat_handle_intern.c"
+#include "gnunet_chat_message.h"
 #include <gnunet/gnunet_arm_service.h>
 #include <gnunet/gnunet_common.h>
 #include <gnunet/gnunet_reclaim_service.h>
+#include <gnunet/gnunet_scheduler_lib.h>
 
 static const unsigned int initial_map_size_of_handle = 8;
 static const unsigned int minimum_amount_of_other_members_in_group = 2;
@@ -95,6 +97,7 @@ handle_create_from_config (const struct GNUNET_CONFIGURATION_Handle* cfg,
   handle->accounts_head = NULL;
   handle->accounts_tail = NULL;
 
+  handle->next = NULL;
   handle->current = NULL;
   handle->monitor = NULL;
 
@@ -110,7 +113,9 @@ handle_create_from_config (const struct GNUNET_CONFIGURATION_Handle* cfg,
   handle->tickets_head = NULL;
   handle->tickets_tail = NULL;
 
-  handle->files = NULL;
+  handle->files = GNUNET_CONTAINER_multihashmap_create(
+    initial_map_size_of_handle, GNUNET_NO);
+  
   handle->contexts = NULL;
   handle->contacts = NULL;
   handle->groups = NULL;
@@ -130,6 +135,16 @@ handle_create_from_config (const struct GNUNET_CONFIGURATION_Handle* cfg,
     handle
   );
 
+  handle->fs = GNUNET_FS_start(
+    handle->cfg, "libgnunetchat",
+    notify_handle_fs_progress, handle,
+    GNUNET_FS_FLAGS_NONE,
+    GNUNET_FS_OPTIONS_END
+  );
+
+  handle->gns = NULL;
+  handle->messenger = NULL;
+
   handle->namestore = GNUNET_NAMESTORE_connect(
     handle->cfg
   );
@@ -138,10 +153,6 @@ handle_create_from_config (const struct GNUNET_CONFIGURATION_Handle* cfg,
     handle->cfg
   );
 
-  handle->fs = NULL;
-  handle->gns = NULL;
-  handle->messenger = NULL;
-
   handle->public_key = NULL;
   handle->user_pointer = NULL;
   return handle;
@@ -174,69 +185,47 @@ handle_destroy (struct GNUNET_CHAT_Handle *handle)
 
   if (handle->shutdown_hook)
     GNUNET_SCHEDULER_cancel(handle->shutdown_hook);
-
   if (handle->destruction)
     GNUNET_SCHEDULER_cancel(handle->destruction);
-
-  if (handle->disconnection)
-    GNUNET_SCHEDULER_cancel(handle->disconnection);
+  if (handle->connection)
+    GNUNET_SCHEDULER_cancel(handle->connection);
+  if (handle->refresh)
+    GNUNET_SCHEDULER_cancel(handle->refresh);
 
   if (handle->monitor)
     GNUNET_NAMESTORE_zone_monitor_stop(handle->monitor);
 
+  handle->connection = NULL;
+
   if (handle->current)
     handle_disconnect(handle);
 
-  if (handle->namestore)
-    GNUNET_NAMESTORE_disconnect(handle->namestore);
-
-  struct GNUNET_CHAT_TicketProcess *tickets;
-  while (handle->tickets_head)
-  {
-    tickets = handle->tickets_head;
-
-    if (tickets->iter)
-      GNUNET_RECLAIM_ticket_iteration_stop(tickets->iter);
-
-    if (tickets->op)
-      GNUNET_RECLAIM_cancel(tickets->op);
-
-    GNUNET_CONTAINER_DLL_remove(
-      handle->tickets_head,
-      handle->tickets_tail,
-      tickets
-    );
-
-    GNUNET_free(tickets);
-  }
+  GNUNET_CONTAINER_multihashmap_iterate(
+    handle->files, it_destroy_handle_files, NULL
+  );
 
   if (handle->reclaim)
     GNUNET_RECLAIM_disconnect(handle->reclaim);
 
-  struct GNUNET_CHAT_InternalAccounts *accounts;
+  if (handle->namestore)
+    GNUNET_NAMESTORE_disconnect(handle->namestore);
 
+  struct GNUNET_CHAT_InternalAccounts *accounts;
   while (handle->accounts_head)
   {
     accounts = handle->accounts_head;
 
-    if (accounts->op)
-      GNUNET_IDENTITY_cancel(accounts->op);
+    internal_accounts_stop_method(accounts);
 
     if (accounts->account)
       account_destroy(accounts->account);
 
-    GNUNET_CONTAINER_DLL_remove(
-      handle->accounts_head,
-      handle->accounts_tail,
-      accounts
-    );
-
-    if (accounts->identifier)
-      GNUNET_free(accounts->identifier);
-
-    GNUNET_free(accounts);
+    internal_accounts_destroy(accounts);
   }
 
+  if (handle->fs)
+    GNUNET_FS_stop(handle->fs);
+
   if (handle->identity)
     GNUNET_IDENTITY_disconnect(handle->identity);
 
@@ -260,6 +249,8 @@ handle_destroy (struct GNUNET_CHAT_Handle *handle)
   if (handle->arm)
     GNUNET_ARM_disconnect(handle->arm);
 
+  GNUNET_CONTAINER_multihashmap_destroy(handle->files);
+
   if (handle->directory)
     GNUNET_free(handle->directory);
 
@@ -288,6 +279,7 @@ handle_update_identity(struct GNUNET_CHAT_Handle *handle)
 {
   GNUNET_assert(
     (handle) &&
+    (handle->current) &&
 		(handle->contexts) &&
 		(handle->groups) &&
 		(handle->contacts)
@@ -304,6 +296,7 @@ handle_update_identity(struct GNUNET_CHAT_Handle *handle)
 
   handle_send_internal_message(
     handle,
+    handle->current,
     NULL,
     GNUNET_CHAT_FLAG_LOGIN,
     NULL
@@ -337,7 +330,7 @@ handle_connect (struct GNUNET_CHAT_Handle *handle,
 		(!(handle->groups)) &&
 		(!(handle->contacts)) &&
 		(!(handle->contexts)) &&
-		(!(handle->files))
+		(handle->files)
   );
 
   if (handle->monitor)
@@ -346,8 +339,6 @@ handle_connect (struct GNUNET_CHAT_Handle *handle,
     handle->monitor = NULL;
   }
 
-  handle->files = GNUNET_CONTAINER_multihashmap_create(
-    initial_map_size_of_handle, GNUNET_NO);
   handle->contexts = GNUNET_CONTAINER_multihashmap_create(
     initial_map_size_of_handle, GNUNET_NO);
   handle->contacts = GNUNET_CONTAINER_multishortmap_create(
@@ -355,37 +346,18 @@ handle_connect (struct GNUNET_CHAT_Handle *handle,
   handle->groups = GNUNET_CONTAINER_multihashmap_create(
     initial_map_size_of_handle, GNUNET_NO);
 
-  const char *name = account->name;
-
-  char* fs_client_name = NULL;
-  GNUNET_asprintf (
-    &fs_client_name,
-    "GNUNET_CHAT_%s%s",
-    name? "_" : "anonymous",
-    name? name : ""
-  );
-
-  handle->fs = GNUNET_FS_start(
-    handle->cfg, fs_client_name,
-    notify_handle_fs_progress, handle,
-    GNUNET_FS_FLAGS_NONE,
-    GNUNET_FS_OPTIONS_END
-  );
-
-  GNUNET_free(fs_client_name);
-
   handle->gns = GNUNET_GNS_connect(handle->cfg);
 
-  const struct GNUNET_CRYPTO_PrivateKey *key = NULL;
-  if (account->ego)
-    key = GNUNET_IDENTITY_ego_get_private_key(account->ego);
+  const struct GNUNET_CRYPTO_PrivateKey *key;
+  key = account_get_key(account);
 
   handle->messenger = GNUNET_MESSENGER_connect(
-    handle->cfg, name, key,
+    handle->cfg, account->name, key,
     on_handle_message,
     handle
   );
 
+  handle->next = NULL;
   handle->current = account;
   handle_update_identity(handle);
 }
@@ -404,11 +376,18 @@ handle_disconnect (struct GNUNET_CHAT_Handle *handle)
 
   handle_send_internal_message(
     handle,
+    handle->current,
     NULL,
     GNUNET_CHAT_FLAG_LOGOUT,
     NULL
   );
 
+  while (handle->attributes_head)
+    internal_attributes_destroy(handle->attributes_head);
+
+  while (handle->tickets_head)
+    internal_tickets_destroy(handle->tickets_head);
+
   GNUNET_CONTAINER_multihashmap_iterate(
     handle->groups, it_destroy_handle_groups, NULL
   );
@@ -471,10 +450,6 @@ handle_disconnect (struct GNUNET_CHAT_Handle *handle)
     handle->files, it_destroy_handle_files, NULL
   );
 
-  if (handle->fs)
-    GNUNET_FS_stop(handle->fs);
-
-  handle->fs = NULL;
   handle->gns = NULL;
   handle->messenger = NULL;
 
@@ -495,40 +470,21 @@ handle_disconnect (struct GNUNET_CHAT_Handle *handle)
     GNUNET_free(lobbies);
   }
 
-  struct GNUNET_CHAT_AttributeProcess *attributes;
-  while (handle->attributes_head)
-  {
-    attributes = handle->attributes_head;
-
-    if (attributes->attribute)
-      GNUNET_free(attributes->attribute);
-
-    if (attributes->iter)
-      GNUNET_RECLAIM_get_attributes_stop(attributes->iter);
-
-    if (attributes->op)
-      GNUNET_RECLAIM_cancel(attributes->op);
-
-    GNUNET_CONTAINER_DLL_remove(
-      handle->attributes_head,
-      handle->attributes_tail,
-      attributes
-    );
-
-    GNUNET_free(attributes);
-  }
-
   GNUNET_CONTAINER_multihashmap_destroy(handle->groups);
   GNUNET_CONTAINER_multishortmap_destroy(handle->contacts);
   GNUNET_CONTAINER_multihashmap_destroy(handle->contexts);
-  GNUNET_CONTAINER_multihashmap_destroy(handle->files);
+  GNUNET_CONTAINER_multihashmap_clear(handle->files);
 
-  handle->files = NULL;
   handle->contexts = NULL;
   handle->contacts = NULL;
   handle->groups = NULL;
 
+  if (handle->connection)
+    GNUNET_SCHEDULER_cancel(handle->connection);
+
   handle->current = NULL;
+  handle->connection = NULL;
+
   handle_update_key(handle);
 }
 
@@ -555,50 +511,34 @@ find_accounts_by_name (struct GNUNET_CHAT_Handle *handle,
   return accounts;
 }
 
-static int
+static enum GNUNET_GenericReturnValue
 update_accounts_operation (struct GNUNET_CHAT_InternalAccounts **out_accounts,
                            struct GNUNET_CHAT_Handle *handle,
                            const char *name,
-                           int wait_for_completion)
+                           enum GNUNET_CHAT_AccountMethod method)
 {
   GNUNET_assert(handle);
 
   struct GNUNET_CHAT_InternalAccounts *accounts = *out_accounts;
 
-  if (!accounts)
+  if (accounts)
+    internal_accounts_stop_method(accounts);
+  else
   {
-    accounts = GNUNET_new(struct GNUNET_CHAT_InternalAccounts);
+    accounts = internal_accounts_create(handle, NULL);
 
     if (!accounts)
       return GNUNET_SYSERR;
-
-    accounts->account = NULL;
-    accounts->handle = handle;
-
-    GNUNET_CONTAINER_DLL_insert_tail(
-	    handle->accounts_head,
-    	handle->accounts_tail,
-    	accounts
-    );
-  }
-  else
-  {
-    if (accounts->identifier)
-      GNUNET_free(accounts->identifier);
-
-    if (accounts->op)
-      GNUNET_IDENTITY_cancel(accounts->op);
   }
 
-  accounts->identifier = name ? GNUNET_strdup(name) : NULL;
-  accounts->wait_for_completion = wait_for_completion;
+  internal_accounts_start_method(accounts, method, name);
 
   *out_accounts = accounts;
 
   return GNUNET_OK;
 }
 
-int
+enum GNUNET_GenericReturnValue
 handle_create_account (struct GNUNET_CHAT_Handle *handle,
 		                   const char *name)
 {
@@ -608,13 +548,14 @@ handle_create_account (struct GNUNET_CHAT_Handle *handle,
   accounts = find_accounts_by_name(handle, name);
 
   if (accounts)
-    return GNUNET_NO;
+    return GNUNET_SYSERR;
 
-  int result = update_accounts_operation(
+  enum GNUNET_GenericReturnValue result;
+  result = update_accounts_operation(
     &accounts, 
     handle, 
-    NULL, 
-    GNUNET_NO
+    name, 
+    GNUNET_CHAT_ACCOUNT_CREATION
   );
 
   if (GNUNET_OK != result)
@@ -635,7 +576,7 @@ handle_create_account (struct GNUNET_CHAT_Handle *handle,
   return result;
 }
 
-int
+enum GNUNET_GenericReturnValue
 handle_delete_account (struct GNUNET_CHAT_Handle *handle,
 		                   const char *name)
 {
@@ -644,11 +585,15 @@ handle_delete_account (struct GNUNET_CHAT_Handle *handle,
   struct GNUNET_CHAT_InternalAccounts *accounts;
   accounts = find_accounts_by_name(handle, name);
 
-  int result = update_accounts_operation(
+  if (!accounts)
+    return GNUNET_SYSERR;
+
+  enum GNUNET_GenericReturnValue result;
+  result = update_accounts_operation(
     &accounts, 
     handle, 
     NULL, 
-    GNUNET_YES
+    GNUNET_CHAT_ACCOUNT_DELETION
   );
 
   if (GNUNET_OK != result)
@@ -667,7 +612,7 @@ handle_delete_account (struct GNUNET_CHAT_Handle *handle,
   return result;
 }
 
-int
+enum GNUNET_GenericReturnValue
 handle_rename_account (struct GNUNET_CHAT_Handle *handle,
                        const char *old_name,
                        const char *new_name)
@@ -677,11 +622,15 @@ handle_rename_account (struct GNUNET_CHAT_Handle *handle,
   struct GNUNET_CHAT_InternalAccounts *accounts;
   accounts = find_accounts_by_name(handle, old_name);
 
-  int result = update_accounts_operation(
+  if (!accounts)
+    return GNUNET_SYSERR;
+
+  enum GNUNET_GenericReturnValue result;
+  result = update_accounts_operation(
     &accounts, 
     handle, 
     NULL, 
-    GNUNET_YES
+    GNUNET_CHAT_ACCOUNT_RENAMING
   );
 
   if (GNUNET_OK != result)
@@ -701,6 +650,52 @@ handle_rename_account (struct GNUNET_CHAT_Handle *handle,
   return result;
 }
 
+enum GNUNET_GenericReturnValue
+handle_delete_lobby (struct GNUNET_CHAT_Handle *handle,
+                     const struct GNUNET_CHAT_Lobby *lobby)
+{
+  GNUNET_assert((handle) && (lobby));
+
+  if (!(lobby->context))
+    return GNUNET_SYSERR;
+
+  const struct GNUNET_HashCode *key = GNUNET_MESSENGER_room_get_key(
+    lobby->context->room
+  );
+
+  if (!key)
+    return GNUNET_SYSERR;
+
+  struct GNUNET_CHAT_InternalAccounts *accounts = NULL;
+  enum GNUNET_GenericReturnValue result;
+  result = update_accounts_operation(
+    &accounts, 
+    handle, 
+    NULL, 
+    GNUNET_CHAT_ACCOUNT_DELETION
+  );
+
+  if (GNUNET_OK != result)
+    return result;
+
+  char *name;
+  util_lobby_name(key, &name);
+
+  accounts->op = GNUNET_IDENTITY_delete(
+    handle->identity,
+    name,
+    cb_lobby_deletion,
+    accounts
+  );
+
+  GNUNET_free(name);
+
+  if (!accounts->op)
+    return GNUNET_SYSERR;
+
+  return result;
+}
+
 const char*
 handle_get_directory (const struct GNUNET_CHAT_Handle *handle)
 {
@@ -715,7 +710,26 @@ handle_get_directory (const struct GNUNET_CHAT_Handle *handle)
     return handle->current->directory;
 }
 
-int
+char*
+handle_create_file_path (const struct GNUNET_CHAT_Handle *handle,
+                         const struct GNUNET_HashCode *hash)
+{
+  GNUNET_assert((handle) && (hash));
+
+  const char *directory = handle_get_directory(handle);
+
+  if (!directory)
+    return NULL;
+
+  char *filename;
+  util_get_filename (
+    directory, "files", hash, &filename
+  );
+
+  return filename;
+}
+
+enum GNUNET_GenericReturnValue
 handle_update (struct GNUNET_CHAT_Handle *handle)
 {
   GNUNET_assert((handle) && (handle->current));
@@ -728,11 +742,15 @@ handle_update (struct GNUNET_CHAT_Handle *handle)
   struct GNUNET_CHAT_InternalAccounts *accounts;
   accounts = find_accounts_by_name(handle, name);
 
-  int result = update_accounts_operation(
+  if (!accounts)
+    return GNUNET_SYSERR;
+
+  enum GNUNET_GenericReturnValue result;
+  result = update_accounts_operation(
     &accounts, 
     handle, 
     name, 
-    GNUNET_YES
+    GNUNET_CHAT_ACCOUNT_UPDATING
   );
 
   if (GNUNET_OK != result)
@@ -756,16 +774,15 @@ handle_get_key (const struct GNUNET_CHAT_Handle *handle)
 {
   GNUNET_assert(handle);
 
-  if ((!(handle->current)) || (!(handle->current->ego)))
+  if (!(handle->current))
     return NULL;
 
-  return GNUNET_IDENTITY_ego_get_private_key(
-    handle->current->ego
-  );
+  return account_get_key(handle->current);
 }
 
 void
 handle_send_internal_message (struct GNUNET_CHAT_Handle *handle,
+                              const struct GNUNET_CHAT_Account *account,
                               struct GNUNET_CHAT_Context *context,
                               enum GNUNET_CHAT_MessageFlag flag,
                               const char *warning)
@@ -780,7 +797,7 @@ handle_send_internal_message (struct GNUNET_CHAT_Handle *handle,
   );
 
   internal->msg = message_create_internally(
-    context, flag, warning
+    account, context, flag, warning
   );
 
   if (!(internal->msg))
@@ -820,6 +837,7 @@ handle_send_room_name (struct GNUNET_CHAT_Handle *handle,
     return;
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
 
   msg.header.kind = GNUNET_MESSENGER_KIND_NAME;
   msg.body.name.name = GNUNET_strdup(name);
@@ -829,7 +847,7 @@ handle_send_room_name (struct GNUNET_CHAT_Handle *handle,
   GNUNET_free(msg.body.name.name);
 }
 
-int
+enum GNUNET_GenericReturnValue
 handle_request_context_by_room (struct GNUNET_CHAT_Handle *handle,
 				                        struct GNUNET_MESSENGER_Room *room)
 {
@@ -847,11 +865,16 @@ handle_request_context_by_room (struct GNUNET_CHAT_Handle *handle,
 
   struct GNUNET_CHAT_CheckHandleRoomMembers check;
 
-  if ((context) && (context->type == GNUNET_CHAT_CONTEXT_TYPE_UNKNOWN))
-    goto check_type;
-  else if (context)
-    return GNUNET_OK;
+  if (!context)
+    goto new_context;
 
+  if ((GNUNET_CHAT_CONTEXT_TYPE_UNKNOWN == context->type) &&
+      (GNUNET_YES != context->deleted))
+    goto check_type;
+  
+  return GNUNET_OK;
+
+new_context:
   context = context_create_from_room(handle, room);
 
   if (GNUNET_OK != GNUNET_CONTAINER_multihashmap_put(
@@ -878,8 +901,7 @@ check_type:
 						      check.contact,
 						      context)))
   {
-    context->deleted = GNUNET_YES;
-    context_write_records(context);
+    context_delete(context, GNUNET_NO);
 
     context->type = GNUNET_CHAT_CONTEXT_TYPE_CONTACT;
     context->deleted = GNUNET_NO;
@@ -888,8 +910,7 @@ check_type:
   }
   else if (checks >= minimum_amount_of_other_members_in_group)
   {
-    context->deleted = GNUNET_YES;
-    context_write_records(context);
+    context_delete(context, GNUNET_NO);
 
     context->type = GNUNET_CHAT_CONTEXT_TYPE_GROUP;
     context->deleted = GNUNET_NO;
@@ -927,8 +948,9 @@ setup_group:
   {
     handle_send_internal_message(
       handle,
+      NULL,
       context,
-      GNUNET_CHAT_FLAG_UPDATE,
+      GNUNET_CHAT_FLAG_UPDATE_CONTEXT,
       NULL
     );
 
@@ -1002,12 +1024,15 @@ handle_process_records (struct GNUNET_CHAT_Handle *handle,
   if (!record)
     return NULL;
 
+  struct GNUNET_HashCode key;
+  GNUNET_memcpy (&key, &(record->key), sizeof(key));
+
   struct GNUNET_CHAT_Context *context = GNUNET_CONTAINER_multihashmap_get(
     handle->contexts,
-    &(record->key)
+    &key
   );
 
-  if (context)
+  if ((context) && (context->room))
   {
     context_read_records(context, label, count, data);
     return NULL;
@@ -1016,11 +1041,17 @@ handle_process_records (struct GNUNET_CHAT_Handle *handle,
   struct GNUNET_MESSENGER_Room *room = GNUNET_MESSENGER_enter_room(
     handle->messenger,
     &(record->door),
-    &(record->key)
+    &key
   );
 
   if (!room)
     return NULL;
+  else if (context)
+  {
+    context_update_room(context, room, GNUNET_NO);
+    context_read_records(context, label, count, data);
+    return NULL;
+  }
 
   context = context_create_from_room(handle, room);
   context_read_records(context, label, count, data);
@@ -1028,7 +1059,7 @@ handle_process_records (struct GNUNET_CHAT_Handle *handle,
   handle_send_room_name(handle, room);
 
   if (GNUNET_OK != GNUNET_CONTAINER_multihashmap_put(
-      handle->contexts, &(record->key), context,
+      handle->contexts, &key, context,
       GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_FAST))
   {
     context_destroy(context);
@@ -1045,51 +1076,9 @@ handle_process_records (struct GNUNET_CHAT_Handle *handle,
     group_publish(group);
 
   if (GNUNET_OK != GNUNET_CONTAINER_multihashmap_put(
-      handle->groups, &(record->key), group,
+      handle->groups, &key, group,
       GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_FAST))
     group_destroy(group);
 
   return context;
 }
-
-void
-handle_update_tickets (struct GNUNET_CHAT_Handle *handle,
-                       const struct GNUNET_CRYPTO_PrivateKey *identity)
-{
-  GNUNET_assert((handle) && (identity));
-
-  if (!handle->reclaim)
-    return;
-
-  struct GNUNET_CHAT_TicketProcess *tickets = GNUNET_new(
-    struct GNUNET_CHAT_TicketProcess
-  );
-
-  if (!tickets)
-    return;
-
-  memset(tickets, 0, sizeof(struct GNUNET_CHAT_TicketProcess));
-  tickets->handle = handle;
-  GNUNET_memcpy(
-    &(tickets->identity),
-    identity,
-    sizeof(tickets->identity)
-  );
-
-  tickets->iter = GNUNET_RECLAIM_ticket_iteration_start(
-    handle->reclaim,
-    identity,
-    cb_task_error_ticket_update,
-    tickets,
-    cb_iterate_ticket_update,
-    tickets,
-    cb_task_finish_ticket_update,
-    tickets
-  );
-
-  GNUNET_CONTAINER_DLL_insert_tail(
-    handle->tickets_head,
-    handle->tickets_tail,
-    tickets
-  );
-}
