@@ -34,6 +34,7 @@
 #include <gnunet/gnunet_namestore_service.h>
 #include <gnunet/gnunet_reclaim_lib.h>
 #include <gnunet/gnunet_reclaim_service.h>
+#include <gnunet/gnunet_time_lib.h>
 #include <gnunet/gnunet_util_lib.h>
 
 #include "gnunet_chat_lib.h"
@@ -42,6 +43,10 @@
 #include "gnunet_chat_message.h"
 #include "gnunet_chat_uri.h"
 
+#include "internal/gnunet_chat_accounts.h"
+#include "internal/gnunet_chat_attribute_process.h"
+#include "internal/gnunet_chat_ticket_process.h"
+
 struct GNUNET_CHAT_Handle;
 
 struct GNUNET_CHAT_InternalServices
@@ -59,20 +64,6 @@ struct GNUNET_CHAT_InternalMessages
   struct GNUNET_CHAT_InternalMessages *prev;
 };
 
-struct GNUNET_CHAT_InternalAccounts
-{
-  struct GNUNET_CHAT_Account *account;
-  char *identifier;
-
-  struct GNUNET_CHAT_Handle *handle;
-  struct GNUNET_IDENTITY_Operation *op;
-
-  int wait_for_completion;
-
-  struct GNUNET_CHAT_InternalAccounts *next;
-  struct GNUNET_CHAT_InternalAccounts *prev;
-};
-
 struct GNUNET_CHAT_InternalLobbies
 {
   struct GNUNET_CHAT_Lobby *lobby;
@@ -91,41 +82,13 @@ struct GNUNET_CHAT_UriLookups
   struct GNUNET_CHAT_UriLookups *prev;
 };
 
-struct GNUNET_CHAT_AttributeProcess
-{
-  struct GNUNET_CHAT_Handle *handle;
-
-  struct GNUNET_RECLAIM_Attribute *attribute;
-
-  GNUNET_CHAT_AttributeCallback callback;
-  void *closure;
-
-  struct GNUNET_RECLAIM_AttributeIterator *iter;
-  struct GNUNET_RECLAIM_Operation *op;
-
-  struct GNUNET_CHAT_AttributeProcess *next;
-  struct GNUNET_CHAT_AttributeProcess *prev;
-};
-
-struct GNUNET_CHAT_TicketProcess
-{
-  struct GNUNET_CHAT_Handle *handle;
-
-  struct GNUNET_CRYPTO_PrivateKey identity;
-
-  struct GNUNET_RECLAIM_TicketIterator *iter;
-  struct GNUNET_RECLAIM_Operation *op;
-
-  struct GNUNET_CHAT_TicketProcess *next;
-  struct GNUNET_CHAT_TicketProcess *prev;
-};
-
 struct GNUNET_CHAT_Handle
 {
   const struct GNUNET_CONFIGURATION_Handle* cfg;
   struct GNUNET_SCHEDULER_Task *shutdown_hook;
   struct GNUNET_SCHEDULER_Task *destruction;
-  struct GNUNET_SCHEDULER_Task *disconnection;
+  struct GNUNET_SCHEDULER_Task *connection;
+  struct GNUNET_SCHEDULER_Task *refresh;
 
   struct GNUNET_CHAT_InternalServices *services_head;
   struct GNUNET_CHAT_InternalServices *services_tail;
@@ -141,6 +104,7 @@ struct GNUNET_CHAT_Handle
   struct GNUNET_CHAT_InternalAccounts *accounts_head;
   struct GNUNET_CHAT_InternalAccounts *accounts_tail;
 
+  const struct GNUNET_CHAT_Account *next;
   const struct GNUNET_CHAT_Account *current;
   struct GNUNET_NAMESTORE_ZoneMonitor *monitor;
 
@@ -233,7 +197,7 @@ handle_disconnect (struct GNUNET_CHAT_Handle *handle);
  * @param[in] name Chat account name
  * @return #GNUNET_OK on success, otherwise #GNUNET_SYSERR
  */
-int
+enum GNUNET_GenericReturnValue
 handle_create_account (struct GNUNET_CHAT_Handle *handle,
 		                   const char *name);
 
@@ -245,7 +209,7 @@ handle_create_account (struct GNUNET_CHAT_Handle *handle,
  * @param[in] name Chat account name
  * @return #GNUNET_OK on success, otherwise #GNUNET_SYSERR
  */
-int
+enum GNUNET_GenericReturnValue
 handle_delete_account (struct GNUNET_CHAT_Handle *handle,
 		                   const char *name);
 
@@ -259,11 +223,23 @@ handle_delete_account (struct GNUNET_CHAT_Handle *handle,
  * @param[in] new_name New chat account name
  * @return #GNUNET_OK on success, otherwise #GNUNET_SYSERR
  */
-int
+enum GNUNET_GenericReturnValue
 handle_rename_account (struct GNUNET_CHAT_Handle *handle,
                        const char *old_name,
                        const char *new_name);
 
+/**
+ * Enqueues a deletion for a chat <i>lobby</i> for a 
+ * given chat <i>handle</i>.
+ *
+ * @param[in,out] handle Chat handle
+ * @param[in] lobby Chat lobby
+ * @return #GNUNET_OK on success, otherwise #GNUNET_SYSERR
+ */
+enum GNUNET_GenericReturnValue
+handle_delete_lobby (struct GNUNET_CHAT_Handle *handle,
+                     const struct GNUNET_CHAT_Lobby *lobby);
+
 /**
  * Returns the main directory path to store information
  * of a given chat <i>handle</i>.
@@ -274,6 +250,19 @@ handle_rename_account (struct GNUNET_CHAT_Handle *handle,
 const char*
 handle_get_directory (const struct GNUNET_CHAT_Handle *handle);
 
+/**
+ * Returns an allocated string providing the full path to
+ * a file stored by a chat <i>handle</i> with a given 
+ * <i>hash</i>.
+ *
+ * @param[in] handle Chat handle
+ * @param[in] hash File hash
+ * @return File path
+ */
+char*
+handle_create_file_path (const struct GNUNET_CHAT_Handle *handle,
+                         const struct GNUNET_HashCode *hash);
+
 /**
  * Updates the used private key by creating a new identity
  * using the same identifier as currently in use, replacing
@@ -282,7 +271,7 @@ handle_get_directory (const struct GNUNET_CHAT_Handle *handle);
  * @param[in] handle Chat handle
  * @return #GNUNET_OK on success, #GNUNET_SYSERR on failure
  */
-int
+enum GNUNET_GenericReturnValue
 handle_update (struct GNUNET_CHAT_Handle *handle);
 
 /**
@@ -297,16 +286,19 @@ handle_get_key (const struct GNUNET_CHAT_Handle *handle);
 
 /**
  * Sends an internal chat message from a given chat
- * <i>handle</i> with an optional chat <i>context</i>,
- * a custom <i>flag</i> and an optional <i>warning</i> text.
+ * <i>handle</i> with an optional chat <i>account</i> or
+ * <i>context</i>, a custom <i>flag</i> and an optional 
+ * <i>warning</i> text.
  *
  * @param[in,out] handle Chat handle
+ * @param[in] account Chat account or NULL
  * @param[in,out] context Chat context or NULL
  * @param[in] flag Chat message flag
  * @param[in] warning Warning text
  */
 void
 handle_send_internal_message (struct GNUNET_CHAT_Handle *handle,
+                              const struct GNUNET_CHAT_Account *account,
                               struct GNUNET_CHAT_Context *context,
                               enum GNUNET_CHAT_MessageFlag flag,
                               const char *warning);
@@ -331,7 +323,7 @@ handle_send_room_name (struct GNUNET_CHAT_Handle *handle,
  * @param[in,out] room Messenger room
  * @return #GNUNET_OK on success, otherwise #GNUNET_SYSERR
  */
-int
+enum GNUNET_GenericReturnValue
 handle_request_context_by_room (struct GNUNET_CHAT_Handle *handle,
 				                        struct GNUNET_MESSENGER_Room *room);
 
@@ -377,15 +369,4 @@ handle_process_records (struct GNUNET_CHAT_Handle *handle,
                         unsigned int count,
                         const struct GNUNET_GNSRECORD_Data *data);
 
-/**
- * Updates the tickets of a given chat <i>handle</i>, revoking 
- * all tickets of a <i>identity</i>.
- *
- * @param[in,out] handle Chat handle
- * @param[in] identity Identity private key
- */
-void
-handle_update_tickets (struct GNUNET_CHAT_Handle *handle,
-                       const struct GNUNET_CRYPTO_PrivateKey *identity);
-
 #endif /* GNUNET_CHAT_HANDLE_H_ */
