@@ -24,14 +24,17 @@
 
 #include "gnunet_chat_contact.h"
 #include "gnunet_chat_context.h"
+#include "gnunet_chat_discourse.h"
 #include "gnunet_chat_file.h"
 #include "gnunet_chat_group.h"
 #include "gnunet_chat_handle.h"
 #include "gnunet_chat_invitation.h"
+#include "gnunet_chat_lobby.h"
 #include "gnunet_chat_message.h"
 #include "gnunet_chat_tagging.h"
 #include "gnunet_chat_ticket.h"
 #include "gnunet_chat_util.h"
+#include "internal/gnunet_chat_accounts.h"
 
 #include <gnunet/gnunet_arm_service.h>
 #include <gnunet/gnunet_common.h>
@@ -254,16 +257,13 @@ notify_handle_fs_progress(void* cls,
       );
 
       file->unindex = NULL;
-      const char *directory = handle_get_directory(chat);
-
-      if (!directory)
-        break;
-
-      char *filename;
-      util_get_filename (
-        directory, "files", &(file->hash), &filename
+      char *filename = handle_create_file_path(
+        chat, &(file->hash)
       );
 
+      if (!filename)
+        break;
+
       if (GNUNET_YES == GNUNET_DISK_file_test_read(filename))
         remove(filename);
 
@@ -277,101 +277,21 @@ notify_handle_fs_progress(void* cls,
   return NULL;
 }
 
-void
-cb_task_finish_ticket_update (void *cls)
+static void
+on_handle_refresh (void *cls)
 {
   GNUNET_assert(cls);
 
-  struct GNUNET_CHAT_TicketProcess *tickets = (
-    (struct GNUNET_CHAT_TicketProcess*) cls
-  );
+  struct GNUNET_CHAT_Handle* handle = cls;
 
-  struct GNUNET_CHAT_Handle *handle = tickets->handle;
-
-  GNUNET_CONTAINER_DLL_remove(
-    handle->tickets_head,
-    handle->tickets_tail,
-    tickets
-  );
-
-  GNUNET_free(tickets);
-}
-
-void
-cb_task_error_ticket_update (void *cls)
-{
-  GNUNET_assert(cls);
-
-  struct GNUNET_CHAT_TicketProcess *tickets = (
-    (struct GNUNET_CHAT_TicketProcess*) cls
-  );
+  handle->refresh = NULL;
 
   handle_send_internal_message(
-    tickets->handle,
+    handle,
     NULL,
-    GNUNET_CHAT_FLAG_WARNING,
-    "Ticket update failed!"
-  );
-
-  cb_task_finish_ticket_update(cls);
-}
-
-static void
-cont_revoke_ticket_with_status (void *cls,
-                                int32_t success,
-                                const char *emsg)
-{
-  GNUNET_assert(cls);
-
-  struct GNUNET_CHAT_TicketProcess *tickets = (
-    (struct GNUNET_CHAT_TicketProcess*) cls
-  );
-
-  tickets->op = NULL;
-
-  struct GNUNET_CHAT_Handle *handle = tickets->handle;
-
-  if (GNUNET_SYSERR == success)
-  {
-    handle_send_internal_message(
-      handle,
-      NULL,
-      GNUNET_CHAT_KIND_WARNING,
-      emsg
-    );
-
-    if (tickets->iter)
-      GNUNET_RECLAIM_ticket_iteration_stop(tickets->iter);
-
-    tickets->iter = NULL;
-    return;
-  }
-
-  if (tickets->iter)
-    GNUNET_RECLAIM_ticket_iteration_next(tickets->iter);
-}
-
-void
-cb_iterate_ticket_update (void *cls,
-                          const struct GNUNET_RECLAIM_Ticket *ticket)
-{
-  GNUNET_assert(cls);
-
-  struct GNUNET_CHAT_TicketProcess *tickets = (
-    (struct GNUNET_CHAT_TicketProcess*) cls
-  );
-
-  struct GNUNET_CHAT_Handle *handle = tickets->handle;
-
-  if (tickets->op)
-    GNUNET_RECLAIM_cancel(tickets->op);
-
-  tickets->op = GNUNET_RECLAIM_ticket_revoke(
-    handle->reclaim,
-    &(tickets->identity),
-    ticket,
-    cont_revoke_ticket_with_status,
-    tickets
+    NULL,
+    GNUNET_CHAT_FLAG_REFRESH,
+    NULL
   );
 }
 
@@ -383,12 +303,12 @@ on_handle_gnunet_identity (void *cls,
 {
   GNUNET_assert(cls);
 
-  struct GNUNET_CHAT_Handle* handle = cls;
-
-  if (!ctx)
+  if ((name) && (GNUNET_YES == util_is_lobby_name(name)))
     return;
 
-  if (!ego)
+  struct GNUNET_CHAT_Handle* handle = cls;
+
+  if ((!ctx) || (!ego))
     goto send_refresh;
 
   struct GNUNET_CHAT_InternalAccounts *accounts = handle->accounts_head;
@@ -405,52 +325,30 @@ on_handle_gnunet_identity (void *cls,
     {
       util_set_name_field(name, &(accounts->account->name));
 
-      if (handle->current == accounts->account)
-	      handle_send_internal_message(
-          handle,
-          NULL,
-          GNUNET_CHAT_FLAG_LOGIN,
-          NULL
-        );
+      handle_send_internal_message(
+        handle,
+        accounts->account,
+        NULL,
+        GNUNET_CHAT_FLAG_UPDATE_ACCOUNT,
+        NULL
+      );
     }
-    else
+    else if (!(accounts->op))
     {
-      const struct GNUNET_CRYPTO_PrivateKey *key;
-      key = GNUNET_IDENTITY_ego_get_private_key(ego);
-
-      if (key)
-        handle_update_tickets(handle, key);
-
       if (handle->current == accounts->account)
 	      handle_disconnect(handle);
 
       account_destroy(accounts->account);
-
-      if (accounts->op)
-      {
-        accounts->account = NULL;
-        goto send_refresh;
-      }
-
-      GNUNET_CONTAINER_DLL_remove(
-        handle->accounts_head,
-        handle->accounts_tail,
-        accounts
-      );
-
-      if (accounts->identifier)
-	      GNUNET_free(accounts->identifier);
-
-      GNUNET_free(accounts);
+      internal_accounts_destroy(accounts);
     }
 
     goto send_refresh;
 
 check_matching_name:
-    if ((name) && (accounts->account->name) &&
+    if ((name) && (accounts->account->name) && (ego) &&
 	      (0 == strcmp(accounts->account->name, name)))
     {
-      accounts->account->ego = ego;
+      account_update_ego(accounts->account, handle, ego);
       goto send_refresh;
     }
 
@@ -461,25 +359,24 @@ skip_account:
   if (!name)
     return;
 
-  accounts = GNUNET_new(struct GNUNET_CHAT_InternalAccounts);
-  accounts->account = account_create_from_ego(ego, name);
-
-  accounts->handle = handle;
-  accounts->op = NULL;
-
-  accounts->wait_for_completion = GNUNET_NO;
+  accounts = internal_accounts_create(
+    handle,
+    account_create_from_ego(ego, name)
+  );
 
   if (handle->directory)
     account_update_directory(accounts->account, handle->directory);
 
-  GNUNET_CONTAINER_DLL_insert_tail(
-    handle->accounts_head,
-    handle->accounts_tail,
-    accounts
-  );
-
 send_refresh:
-  handle_send_internal_message(handle, NULL, GNUNET_CHAT_FLAG_REFRESH, NULL);
+  if (handle->refresh)
+    return;
+  
+  handle->refresh = GNUNET_SCHEDULER_add_at_with_priority(
+    GNUNET_TIME_absolute_get(),
+    GNUNET_SCHEDULER_PRIORITY_IDLE,
+    on_handle_refresh,
+    handle
+  );
 }
 
 void
@@ -493,32 +390,25 @@ cb_account_creation (void *cls,
     (struct GNUNET_CHAT_InternalAccounts*) cls
   );
 
-  struct GNUNET_CHAT_Handle *handle = accounts->handle;
+  accounts->op = NULL;
 
-  GNUNET_CONTAINER_DLL_remove(
-    handle->accounts_head,
-    handle->accounts_tail,
-    accounts
-  );
-
-  if (accounts->identifier)
-    GNUNET_free(accounts->identifier);
-
-  GNUNET_free(accounts);
-
-  if (GNUNET_EC_NONE != ec)
-  {
-    handle_send_internal_message(
-	    handle,
-      NULL,
-      GNUNET_CHAT_FLAG_WARNING,
-      GNUNET_ErrorCode_get_hint(ec)
+  if ((!(accounts->account)) && (accounts->identifier))
+    accounts->account = account_create(
+      accounts->identifier
     );
-
+  
+  internal_accounts_stop_method(accounts);
+  
+  if (GNUNET_EC_NONE == ec)
     return;
-  }
-  else if (key)
-    handle_send_internal_message(handle, NULL, GNUNET_CHAT_FLAG_REFRESH, NULL);
+
+  handle_send_internal_message(
+    accounts->handle,
+    accounts->account,
+    NULL,
+    GNUNET_CHAT_FLAG_WARNING,
+    GNUNET_ErrorCode_get_hint(ec)
+  );
 }
 
 void
@@ -531,30 +421,36 @@ cb_account_deletion (void *cls,
     (struct GNUNET_CHAT_InternalAccounts*) cls
   );
 
-  struct GNUNET_CHAT_Handle *handle = accounts->handle;
+  accounts->op = NULL;
 
-  GNUNET_CONTAINER_DLL_remove(
-    handle->accounts_head,
-    handle->accounts_tail,
-    accounts
-  );
+  internal_accounts_stop_method(accounts);
 
-  if (accounts->identifier)
-    GNUNET_free(accounts->identifier);
-
-  GNUNET_free(accounts);
+  if (accounts->handle->current == accounts->account)
+	  handle_disconnect(accounts->handle);
 
   if (GNUNET_EC_NONE != ec)
-  {
     handle_send_internal_message(
-      handle,
+      accounts->handle,
+      accounts->account,
       NULL,
       GNUNET_CHAT_FLAG_WARNING,
       GNUNET_ErrorCode_get_hint(ec)
     );
+  else
+  {
+    handle_send_internal_message(
+      accounts->handle,
+      accounts->account,
+      NULL,
+      GNUNET_CHAT_FLAG_DELETE_ACCOUNT,
+      NULL
+    );
 
-    return;
+    account_delete(accounts->account);
   }
+
+  account_destroy(accounts->account);
+  internal_accounts_destroy(accounts);
 }
 
 void
@@ -567,30 +463,46 @@ cb_account_rename (void *cls,
     (struct GNUNET_CHAT_InternalAccounts*) cls
   );
 
-  struct GNUNET_CHAT_Handle *handle = accounts->handle;
+  accounts->op = NULL;
 
-  GNUNET_CONTAINER_DLL_remove(
-    handle->accounts_head,
-    handle->accounts_tail,
-    accounts
+  internal_accounts_stop_method(accounts);
+
+  if (GNUNET_EC_NONE == ec)
+    return;
+
+  handle_send_internal_message(
+    accounts->handle,
+    accounts->account,
+    NULL,
+    GNUNET_CHAT_FLAG_WARNING,
+    GNUNET_ErrorCode_get_hint(ec)
+  );
+}
+
+void
+cb_lobby_deletion (void *cls,
+		               enum GNUNET_ErrorCode ec)
+{
+  GNUNET_assert(cls);
+
+  struct GNUNET_CHAT_InternalAccounts *accounts = (
+    (struct GNUNET_CHAT_InternalAccounts*) cls
   );
 
-  if (accounts->identifier)
-    GNUNET_free(accounts->identifier);
+  accounts->op = NULL;
 
-  GNUNET_free(accounts);
+  internal_accounts_stop_method(accounts);
 
   if (GNUNET_EC_NONE != ec)
-  {
     handle_send_internal_message(
-      handle,
+      accounts->handle,
+      accounts->account,
       NULL,
       GNUNET_CHAT_FLAG_WARNING,
       GNUNET_ErrorCode_get_hint(ec)
     );
 
-    return;
-  }
+  internal_accounts_destroy(accounts);
 }
 
 static void
@@ -604,17 +516,7 @@ cb_account_update_completion (void *cls,
     (struct GNUNET_CHAT_InternalAccounts*) cls
   );
 
-  struct GNUNET_CHAT_Handle *handle = accounts->handle;
-
-  if ((GNUNET_EC_NONE == ec) && (key))
-  {
-    const struct GNUNET_CRYPTO_PrivateKey *prev= handle_get_key(handle);
-
-    if (prev)
-      handle_update_tickets(handle, prev);
-
-    GNUNET_MESSENGER_set_key(handle->messenger, key);
-  }
+  accounts->op = NULL;
 
   cb_account_creation(cls, key, ec);
 }
@@ -629,16 +531,14 @@ cb_account_update (void *cls,
     (struct GNUNET_CHAT_InternalAccounts*) cls
   );
 
-  struct GNUNET_CHAT_Handle *handle = accounts->handle;
-
-  if ((GNUNET_EC_NONE != ec) || (!accounts->identifier))
+  if ((GNUNET_EC_NONE != ec) || (!(accounts->identifier)))
   {
     cb_account_deletion(cls, ec);
     return;
   }
-
+  
   accounts->op = GNUNET_IDENTITY_create(
-    handle->identity,
+    accounts->handle->identity,
     accounts->identifier,
     NULL,
     GNUNET_PUBLIC_KEY_TYPE_ECDSA,
@@ -953,7 +853,21 @@ skip_msg_handing:
     }
     case GNUNET_MESSENGER_KIND_TICKET:
     {
-      struct GNUNET_CHAT_InternalTickets *tickets = GNUNET_new(
+      struct GNUNET_CHAT_InternalTickets *tickets = contact->tickets_head;
+      while (tickets)
+      {
+        if (0 == strncmp(tickets->ticket->ticket.gns_name, 
+                         message->msg->body.ticket.identifier,
+                         sizeof(tickets->ticket->ticket.gns_name)))
+          break;
+
+        tickets = tickets->next;
+      }
+
+      if (tickets)
+        break;
+      
+      tickets = GNUNET_new(
         struct GNUNET_CHAT_InternalTickets
       );
 
@@ -961,7 +875,7 @@ skip_msg_handing:
         break;
 
       tickets->ticket = ticket_create_from_message(
-	      handle, sender, &(message->msg->body.ticket)
+	      handle, contact, &(message->msg->body.ticket)
       );
 
       if (!tickets->ticket)
@@ -977,6 +891,42 @@ skip_msg_handing:
       );
       break;
     }
+    case GNUNET_MESSENGER_KIND_SUBSCRIBE:
+    {
+      const struct GNUNET_ShortHashCode *id = &(message->msg->body.subscribe.discourse);
+      struct GNUNET_CHAT_Discourse *discourse = GNUNET_CONTAINER_multishortmap_get(
+        context->discourses, id
+      );
+
+      if (! discourse)
+      {
+        discourse = discourse_create(context, id);
+
+        if (GNUNET_OK != GNUNET_CONTAINER_multishortmap_put(context->discourses,
+          id, discourse, GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_FAST))
+        {
+          discourse_destroy(discourse);
+          break;
+        }
+      }
+
+      if (GNUNET_MESSENGER_FLAG_SUBSCRIPTION_UNSUBSCRIBE & message->msg->body.subscribe.flags)
+        discourse_unsubscribe(
+          discourse,
+          contact,
+          GNUNET_TIME_absolute_ntoh(message->msg->header.timestamp),
+          GNUNET_TIME_relative_ntoh(message->msg->body.subscribe.time)
+        );
+      else
+        discourse_subscribe(
+          discourse,
+          contact,
+          GNUNET_TIME_absolute_ntoh(message->msg->header.timestamp),
+          GNUNET_TIME_relative_ntoh(message->msg->body.subscribe.time)
+        );
+
+      break;
+    }
     default:
       break;
   }
