@@ -25,19 +25,23 @@
 #include "gnunet_chat_lib.h"
 
 #include <gnunet/gnunet_common.h>
+#include <gnunet/gnunet_fs_service.h>
 #include <gnunet/gnunet_messenger_service.h>
 #include <gnunet/gnunet_reclaim_lib.h>
 #include <gnunet/gnunet_reclaim_service.h>
 #include <gnunet/gnunet_scheduler_lib.h>
 #include <gnunet/gnunet_time_lib.h>
+#include <gnunet/gnunet_util_lib.h>
 #include <libgen.h>
-#include <limits.h>
+#include <stdint.h>
+#include <string.h>
 #include <strings.h>
 
 #define _(String) ((const char*) String)
 
 #include "gnunet_chat_contact.h"
 #include "gnunet_chat_context.h"
+#include "gnunet_chat_discourse.h"
 #include "gnunet_chat_file.h"
 #include "gnunet_chat_group.h"
 #include "gnunet_chat_handle.h"
@@ -165,16 +169,29 @@ GNUNET_CHAT_connect (struct GNUNET_CHAT_Handle *handle,
   if ((!handle) || (handle->destruction))
     return;
 
+  if (handle->connection)
+    GNUNET_SCHEDULER_cancel(handle->connection);
+
   if (handle->current == account)
+  {
+    handle->next = NULL;
+    handle->connection = NULL;
     return;
+  }
 
   if (handle->current)
-    handle_disconnect(handle);
-
-  if (!account)
+  {
+    handle->next = account;
+    handle->connection = NULL;
+    GNUNET_CHAT_disconnect(handle);
     return;
+  }
 
-  handle_connect(handle, account);
+  handle->next = account;
+  handle->connection = GNUNET_SCHEDULER_add_now(
+    task_handle_connection,
+    handle
+  );
 }
 
 
@@ -183,11 +200,20 @@ GNUNET_CHAT_disconnect (struct GNUNET_CHAT_Handle *handle)
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
-  if ((!handle) || (handle->destruction) || 
-      (!(handle->current)) || (handle->disconnection))
+  if ((!handle) || (handle->destruction))
     return;
+  
+  if (handle->connection)
+    GNUNET_SCHEDULER_cancel(handle->connection);
 
-  handle->disconnection = GNUNET_SCHEDULER_add_now(
+  if (!(handle->current))
+  {
+    handle->next = NULL;
+    handle->connection = NULL;
+    return;
+  }
+
+  handle->connection = GNUNET_SCHEDULER_add_now(
     task_handle_disconnection,
     handle
   );
@@ -276,7 +302,7 @@ void
 GNUNET_CHAT_set_attribute (struct GNUNET_CHAT_Handle *handle,
                            const char *name,
                            const char *value,
-                           const struct GNUNET_TIME_Relative *expires)
+                           struct GNUNET_TIME_Relative expires)
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
@@ -290,61 +316,40 @@ GNUNET_CHAT_set_attribute (struct GNUNET_CHAT_Handle *handle,
   if ((!key) || (!name))
     return;
 
-  struct GNUNET_CHAT_AttributeProcess *attributes = GNUNET_new(
-    struct GNUNET_CHAT_AttributeProcess
-  );
+  struct GNUNET_CHAT_AttributeProcess *attributes;
+  attributes = internal_attributes_create_store(handle, name, expires);
 
-  memset(attributes, 0, sizeof(struct GNUNET_CHAT_AttributeProcess));
-
-  attributes->handle = handle;
-  attributes->attribute = GNUNET_RECLAIM_attribute_new(
-    name,
-    NULL,
-    GNUNET_RECLAIM_ATTRIBUTE_TYPE_NONE,
-    NULL,
-    0
-  );
-
-  if (!attributes->attribute)
-  {
-    GNUNET_free(attributes);
+  if (!attributes)
     return;
-  }
 
   if (value)
   {
-    void *data = NULL;
-
     enum GNUNET_GenericReturnValue result;
     result = GNUNET_RECLAIM_attribute_string_to_value(
       GNUNET_RECLAIM_ATTRIBUTE_TYPE_STRING,
       value,
-      &data,
+      &(attributes->data),
       &(attributes->attribute->data_size)
     );
 
     if (GNUNET_OK != result)
     {
-      GNUNET_free(attributes->attribute);
-      GNUNET_free(attributes);
+      internal_attributes_destroy(attributes);
       return;
     }
 
-    attributes->attribute->data = data;
+    attributes->attribute->type = GNUNET_RECLAIM_ATTRIBUTE_TYPE_STRING;
+    attributes->attribute->data = attributes->data;
   }
 
-  attributes->op = GNUNET_RECLAIM_attribute_store(
+  attributes->iter = GNUNET_RECLAIM_get_attributes_start(
     handle->reclaim,
     key,
-    attributes->attribute,
-    expires,
-    cont_update_attribute_with_status,
-    attributes
-  );
-
-  GNUNET_CONTAINER_DLL_insert_tail(
-    handle->attributes_head,
-    handle->attributes_tail,
+    cb_task_error_iterate_attribute,
+    attributes,
+    cb_store_attribute,
+    attributes,
+    cb_task_finish_iterate_attribute,
     attributes
   );
 }
@@ -366,41 +371,20 @@ GNUNET_CHAT_delete_attribute (struct GNUNET_CHAT_Handle *handle,
   if ((!key) || (!name))
     return;
 
-  struct GNUNET_CHAT_AttributeProcess *attributes = GNUNET_new(
-    struct GNUNET_CHAT_AttributeProcess
-  );
+  struct GNUNET_CHAT_AttributeProcess *attributes;
+  attributes = internal_attributes_create(handle, name);
 
   if (!attributes)
     return;
 
-  memset(attributes, 0, sizeof(struct GNUNET_CHAT_AttributeProcess));
-
-  attributes->handle = handle;
-  attributes->attribute = GNUNET_RECLAIM_attribute_new(
-    name,
-    NULL,
-    GNUNET_RECLAIM_ATTRIBUTE_TYPE_NONE,
-    NULL,
-    0
-  );
-
-  if (!attributes->attribute)
-  {
-    GNUNET_free(attributes);
-    return;
-  }
-
-  attributes->op = GNUNET_RECLAIM_attribute_delete(
+  attributes->iter = GNUNET_RECLAIM_get_attributes_start(
     handle->reclaim,
     key,
-    attributes->attribute,
-    cont_update_attribute_with_status,
-    attributes
-  );
-
-  GNUNET_CONTAINER_DLL_insert_tail(
-    handle->attributes_head,
-    handle->attributes_tail,
+    cb_task_error_iterate_attribute,
+    attributes,
+    cb_delete_attribute,
+    attributes,
+    cb_task_finish_iterate_attribute,
     attributes
   );
 }
@@ -423,17 +407,12 @@ GNUNET_CHAT_get_attributes (struct GNUNET_CHAT_Handle *handle,
   if (!key)
     return;
 
-  struct GNUNET_CHAT_AttributeProcess *attributes = GNUNET_new(
-    struct GNUNET_CHAT_AttributeProcess
-  );
+  struct GNUNET_CHAT_AttributeProcess *attributes;
+  attributes = internal_attributes_create(handle, NULL);
 
   if (!attributes)
     return;
 
-  memset(attributes, 0, sizeof(struct GNUNET_CHAT_AttributeProcess));
-
-  attributes->handle = handle;
-
   attributes->callback = callback;
   attributes->closure = cls;
 
@@ -447,15 +426,125 @@ GNUNET_CHAT_get_attributes (struct GNUNET_CHAT_Handle *handle,
     cb_task_finish_iterate_attribute,
     attributes
   );
+}
 
-  GNUNET_CONTAINER_DLL_insert_tail(
-    handle->attributes_head,
-    handle->attributes_tail,
+
+void
+GNUNET_CHAT_share_attribute_with (struct GNUNET_CHAT_Handle *handle,
+                                  struct GNUNET_CHAT_Contact *contact,
+                                  const char *name)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!handle) || (handle->destruction) || (!contact))
+    return;
+
+  const struct GNUNET_CRYPTO_PrivateKey *key = handle_get_key(
+    handle
+  );
+
+  const struct GNUNET_CRYPTO_PublicKey *pubkey = contact_get_key(
+    contact
+  );
+
+  if ((!key) || (!pubkey) || (!name))
+    return;
+
+  struct GNUNET_CHAT_AttributeProcess *attributes;
+  attributes = internal_attributes_create_share(handle, contact, name);
+
+  if (!attributes)
+    return;
+
+  attributes->iter = GNUNET_RECLAIM_get_attributes_start(
+    handle->reclaim,
+    key,
+    cb_task_error_iterate_attribute,
+    attributes,
+    cb_share_attribute,
+    attributes,
+    cb_task_finish_iterate_attribute,
     attributes
   );
 }
 
 
+void
+GNUNET_CHAT_unshare_attribute_from (struct GNUNET_CHAT_Handle *handle,
+                                    struct GNUNET_CHAT_Contact *contact,
+                                    const char *name)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!handle) || (handle->destruction) || (!contact))
+    return;
+
+  const struct GNUNET_CRYPTO_PrivateKey *key = handle_get_key(
+    handle
+  );
+
+  if ((!key) || (!name))
+    return;
+
+  struct GNUNET_CHAT_TicketProcess *tickets;
+  tickets = internal_tickets_create(handle, contact, name);
+
+  if (!tickets)
+    return;
+
+  tickets->iter = GNUNET_RECLAIM_ticket_iteration_start(
+    handle->reclaim,
+    key,
+    cb_task_error_iterate_ticket,
+    tickets,
+    cb_iterate_ticket_check,
+    tickets,
+    cb_task_finish_iterate_ticket,
+    tickets
+  );
+}
+
+
+void
+GNUNET_CHAT_get_shared_attributes (struct GNUNET_CHAT_Handle *handle,
+                                   struct GNUNET_CHAT_Contact *contact,
+                                   GNUNET_CHAT_ContactAttributeCallback callback,
+                                   void *cls)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!handle) || (handle->destruction) || (!contact))
+    return;
+
+  const struct GNUNET_CRYPTO_PrivateKey *key = handle_get_key(
+    handle
+  );
+
+  if (!key)
+    return;
+
+  struct GNUNET_CHAT_TicketProcess *tickets;
+  tickets = internal_tickets_create(handle, contact, NULL);
+
+  if (!tickets)
+    return;
+
+  tickets->callback = callback;
+  tickets->closure = cls;
+
+  tickets->iter = GNUNET_RECLAIM_ticket_iteration_start(
+    handle->reclaim,
+    key,
+    cb_task_error_iterate_ticket,
+    tickets,
+    cb_iterate_ticket,
+    tickets,
+    cb_task_finish_iterate_ticket,
+    tickets
+  );
+}
+
+
 struct GNUNET_CHAT_Uri*
 GNUNET_CHAT_uri_parse (const char *uri,
 		                   char **emsg)
@@ -465,44 +554,7 @@ GNUNET_CHAT_uri_parse (const char *uri,
   if (!uri)
     return NULL;
 
-  const size_t prefix_len = strlen(GNUNET_CHAT_URI_PREFIX);
-
-  if (0 != strncasecmp(GNUNET_CHAT_URI_PREFIX, uri, prefix_len))
-  {
-    if (emsg)
-      *emsg = GNUNET_strdup (_ ("CHAT URI malformed (invalid prefix)"));
-
-    return NULL;
-  }
-
-  struct GNUNET_CRYPTO_PublicKey zone;
-
-  const char *data = uri + prefix_len;
-  char *end = strchr(data, '.');
-
-  if (!end)
-  {
-    if (emsg)
-      *emsg = GNUNET_strdup (_ ("CHAT URI malformed (zone key missing)"));
-
-    return NULL;
-  }
-
-  char *zone_data = GNUNET_strndup(data, (size_t) (end - data));
-
-  if (GNUNET_OK != GNUNET_CRYPTO_public_key_from_string(zone_data, &zone))
-  {
-    GNUNET_free(zone_data);
-
-    if (emsg)
-      *emsg = GNUNET_strdup (_ ("CHAT URI malformed (zone key invalid)"));
-
-    return NULL;
-  }
-
-  GNUNET_free(zone_data);
-
-  return uri_create(&zone, end + 1);
+  return uri_parse_from_string(uri, emsg);
 }
 
 
@@ -514,18 +566,19 @@ GNUNET_CHAT_uri_to_string (const struct GNUNET_CHAT_Uri *uri)
   if (!uri)
     return NULL;
 
-  char *key_string = GNUNET_CRYPTO_public_key_to_string(&(uri->zone));
+  return uri_to_string(uri);
+}
 
-  char *string;
-  GNUNET_asprintf (
-    &string,
-    "gnunet://chat/%s.%s",
-    key_string,
-    uri->label
-  );
 
-  GNUNET_free(key_string);
-  return string;
+enum GNUNET_CHAT_UriType
+GNUNET_CHAT_uri_get_type (const struct GNUNET_CHAT_Uri *uri)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if (!uri)
+    return GNUNET_CHAT_URI_TYPE_UNKNOWN;
+
+  return uri->type;
 }
 
 
@@ -607,7 +660,8 @@ GNUNET_CHAT_lobby_join (struct GNUNET_CHAT_Handle *handle,
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
-  if ((!handle) || (handle->destruction) || (!uri) || (!(handle->gns)))
+  if ((!handle) || (handle->destruction) || (!(handle->gns)) ||
+      (!uri) || (GNUNET_CHAT_URI_TYPE_CHAT != uri->type))
     return;
 
   struct GNUNET_CHAT_UriLookups *lookups = GNUNET_new(
@@ -615,12 +669,15 @@ GNUNET_CHAT_lobby_join (struct GNUNET_CHAT_Handle *handle,
   );
 
   lookups->handle = handle;
-  lookups->uri = uri_create(&(uri->zone), uri->label);
+  lookups->uri = uri_create_chat(
+    &(uri->chat.zone),
+    uri->chat.label
+  );
 
   lookups->request = GNUNET_GNS_lookup(
     handle->gns,
-    lookups->uri->label,
-    &(uri->zone),
+    lookups->uri->chat.label,
+    &(uri->chat.zone),
     GNUNET_GNSRECORD_TYPE_MESSENGER_ROOM_ENTRY,
     GNUNET_GNS_LO_DEFAULT,
     cb_lobby_lookup,
@@ -635,6 +692,166 @@ GNUNET_CHAT_lobby_join (struct GNUNET_CHAT_Handle *handle,
 }
 
 
+struct GNUNET_CHAT_File*
+GNUNET_CHAT_request_file (struct GNUNET_CHAT_Handle *handle,
+                          const struct GNUNET_CHAT_Uri *uri)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!handle) || (handle->destruction) || 
+      (!uri) || (GNUNET_CHAT_URI_TYPE_FS != uri->type))
+    return NULL;
+  
+  if (!GNUNET_FS_uri_test_chk(uri->fs.uri))
+    return NULL;
+
+  const struct GNUNET_HashCode *hash = GNUNET_FS_uri_chk_get_file_hash(
+    uri->fs.uri
+  );
+
+  if (!hash)
+    return NULL;
+
+  struct GNUNET_CHAT_File *file = GNUNET_CONTAINER_multihashmap_get(
+    handle->files,
+    hash
+  );
+
+  if (file)
+    return file;
+
+  file = file_create_from_chk_uri(handle, uri->fs.uri);
+
+  if (!file)
+    return NULL;
+
+  if (GNUNET_OK != GNUNET_CONTAINER_multihashmap_put(handle->files, hash, file, 
+      GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_FAST))
+  {
+    file_destroy(file);
+    file = NULL;
+  }
+  
+  return file;
+}
+
+
+struct GNUNET_CHAT_File*
+GNUNET_CHAT_upload_file (struct GNUNET_CHAT_Handle *handle,
+                         const char *path,
+                         GNUNET_CHAT_FileUploadCallback callback,
+                         void *cls)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!handle) || (handle->destruction) ||
+      (!path))
+    return NULL;
+  
+  struct GNUNET_HashCode hash;
+  if (GNUNET_OK != util_hash_file(path, &hash))
+    return NULL;
+
+  char *filename = handle_create_file_path(
+    handle, &hash
+  );
+
+  if (!filename)
+    return NULL;
+
+  struct GNUNET_CHAT_File *file = GNUNET_CONTAINER_multihashmap_get(
+    handle->files,
+    &hash
+  );
+
+  if (file)
+    goto file_binding;
+
+  if ((GNUNET_YES == GNUNET_DISK_file_test(filename)) ||
+      (GNUNET_OK != GNUNET_DISK_directory_create_for_file(filename)) ||
+      (GNUNET_OK != GNUNET_DISK_file_copy(path, filename)))
+  {
+    GNUNET_free(filename);
+    return NULL;
+  }
+
+  char* p = GNUNET_strdup(path);
+
+  file = file_create_from_disk(
+    handle,
+    basename(p),
+    &hash,
+    NULL
+  );
+
+  GNUNET_free(p);
+
+  if (GNUNET_OK != GNUNET_CONTAINER_multihashmap_put(
+      handle->files, &hash, file,
+      GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_FAST))
+  {
+    file_destroy(file);
+    GNUNET_free(filename);
+    return NULL;
+  }
+
+  struct GNUNET_FS_BlockOptions bo;
+
+  bo.anonymity_level = block_anonymity_level;
+  bo.content_priority = block_content_priority;
+  bo.replication_level = block_replication_level;
+  bo.expiration_time = GNUNET_TIME_absolute_get_forever_();
+
+  struct GNUNET_FS_FileInformation* fi = GNUNET_FS_file_information_create_from_file(
+    handle->fs,
+    file,
+    filename,
+    NULL,
+    file->meta,
+    GNUNET_YES,
+    &bo
+  );
+
+  file->publish = GNUNET_FS_publish_start(
+    handle->fs, fi,
+    NULL, NULL, NULL,
+    GNUNET_FS_PUBLISH_OPTION_NONE
+  );
+
+  if (file->publish)
+    file->status |= GNUNET_CHAT_FILE_STATUS_PUBLISH;
+
+  GNUNET_free(filename);
+
+file_binding:
+  file_bind_upload(file, NULL, callback, cls);
+  return file;
+}
+
+
+int
+GNUNET_CHAT_iterate_files (struct GNUNET_CHAT_Handle *handle,
+                           GNUNET_CHAT_FileCallback callback,
+                           void *cls)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!handle) || (handle->destruction))
+    return GNUNET_SYSERR;
+
+  struct GNUNET_CHAT_IterateFiles it;
+  it.handle = handle;
+  it.cb = callback;
+  it.cls = cls;
+
+  return GNUNET_CONTAINER_multihashmap_iterate(
+    handle->files,
+    it_iterate_files,
+    &it
+  );
+}
+
+
 void
 GNUNET_CHAT_set_user_pointer (struct GNUNET_CHAT_Handle *handle,
 			                        void *user_pointer)
@@ -693,6 +910,46 @@ GNUNET_CHAT_account_get_name (const struct GNUNET_CHAT_Account *account)
 }
 
 
+void
+GNUNET_CHAT_account_get_attributes (struct GNUNET_CHAT_Handle *handle,
+                                    const struct GNUNET_CHAT_Account *account,
+                                    GNUNET_CHAT_AccountAttributeCallback callback,
+                                    void *cls)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!handle) || (handle->destruction) || (!account))
+    return;
+
+  const struct GNUNET_CRYPTO_PrivateKey *key = account_get_key(
+    account
+  );
+
+  if (!key)
+    return;
+
+  struct GNUNET_CHAT_AttributeProcess *attributes;
+  attributes = internal_attributes_create_request(handle, account);
+
+  if (!attributes)
+    return;
+
+  attributes->account_callback = callback;
+  attributes->closure = cls;
+
+  attributes->iter = GNUNET_RECLAIM_get_attributes_start(
+    handle->reclaim,
+    key,
+    cb_task_error_iterate_attribute,
+    attributes,
+    cb_iterate_attribute,
+    attributes,
+    cb_task_finish_iterate_attribute,
+    attributes
+  );
+}
+
+
 void
 GNUNET_CHAT_account_set_user_pointer (struct GNUNET_CHAT_Account *account,
 				                              void *user_pointer)
@@ -825,11 +1082,8 @@ GNUNET_CHAT_contact_delete (struct GNUNET_CHAT_Contact *contact)
     contact->handle->contexts, key, contact->context
   );
 
-  GNUNET_MESSENGER_close_room(contact->context->room);
-
-  contact->context->deleted = GNUNET_YES;
-  context_write_records(contact->context);
-
+  context_delete(contact->context, GNUNET_YES);
+  
   context_destroy(contact->context);
   contact_destroy(contact);
   return GNUNET_OK;
@@ -892,7 +1146,10 @@ GNUNET_CHAT_contact_get_context (struct GNUNET_CHAT_Contact *contact)
   if (contact->context)
     return contact->context;
 
-  struct GNUNET_CHAT_Context *context = contact_find_context(contact);
+  struct GNUNET_CHAT_Context *context = contact_find_context(
+    contact,
+    GNUNET_NO
+  );
 
   if ((context) && (GNUNET_CHAT_CONTEXT_TYPE_CONTACT == context->type))
     goto attach_return;
@@ -1041,32 +1298,29 @@ GNUNET_CHAT_contact_is_tagged (const struct GNUNET_CHAT_Contact *contact,
 }
 
 
-int
-GNUNET_CHAT_contact_iterate_tickets (const struct GNUNET_CHAT_Contact *contact,
-                                     GNUNET_CHAT_ContactTicketCallback callback,
-                                     void *cls)
+void
+GNUNET_CHAT_contact_get_attributes (struct GNUNET_CHAT_Contact *contact,
+                                    GNUNET_CHAT_ContactAttributeCallback callback,
+                                    void *cls)
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
   if (!contact)
-    return GNUNET_SYSERR;
+    return;
 
   struct GNUNET_CHAT_InternalTickets *tickets;
-  int result = 0;
-
   tickets = contact->tickets_head;
   
   while (tickets)
   {
-    result++;
-
-    if ((callback) && (GNUNET_NO == callback(cls, contact, tickets->ticket)))
-      break;
+    ticket_consume(
+      tickets->ticket,
+      callback,
+      cls
+    );
 
     tickets = tickets->next;
   }
-
-  return result;
 }
 
 
@@ -1078,6 +1332,7 @@ GNUNET_CHAT_group_leave (struct GNUNET_CHAT_Group *group)
   if ((!group) || (group->destruction))
     return GNUNET_SYSERR;
 
+  group->context->deleted = GNUNET_YES;
   group->destruction = GNUNET_SCHEDULER_add_now(
     task_group_destruction,
     group
@@ -1152,7 +1407,10 @@ GNUNET_CHAT_group_invite_contact (const struct GNUNET_CHAT_Group *group,
   if ((!group) || (!contact) || (!contact->member))
     return;
 
-  struct GNUNET_CHAT_Context *context = contact_find_context(contact);
+  struct GNUNET_CHAT_Context *context = contact_find_context(
+    contact,
+    GNUNET_YES
+  );
 
   if (!context)
     return;
@@ -1166,6 +1424,8 @@ GNUNET_CHAT_group_invite_contact (const struct GNUNET_CHAT_Group *group,
   ));
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
   msg.header.kind = GNUNET_MESSENGER_KIND_INVITE;
   GNUNET_CRYPTO_get_peer_identity(group->handle->cfg, &(msg.body.invite.door));
   GNUNET_memcpy(&(msg.body.invite.key), key, sizeof(msg.body.invite.key));
@@ -1286,9 +1546,12 @@ GNUNET_CHAT_context_request (struct GNUNET_CHAT_Context *context)
 
   context->type = GNUNET_CHAT_CONTEXT_TYPE_CONTACT;
 
-  struct GNUNET_CHAT_Context *other = contact_find_context(contact);
+  struct GNUNET_CHAT_Context *other = contact_find_context(
+    contact,
+    GNUNET_YES
+  );
 
-  if ((!other) || (!(other->room)))
+  if (!other)
     goto cleanup_contact;
 
   struct GNUNET_HashCode key;
@@ -1305,7 +1568,7 @@ GNUNET_CHAT_context_request (struct GNUNET_CHAT_Context *context)
   if (!room)
     goto cleanup_contact;
 
-  context_update_room(context, room);
+  context_update_room(context, room, GNUNET_YES);
 
   if (GNUNET_OK != GNUNET_CONTAINER_multihashmap_put(
       handle->contexts, &key, context,
@@ -1315,6 +1578,8 @@ GNUNET_CHAT_context_request (struct GNUNET_CHAT_Context *context)
   handle_send_room_name(handle, room);
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
   msg.header.kind = GNUNET_MESSENGER_KIND_INVITE;
   GNUNET_CRYPTO_get_peer_identity(handle->cfg, &(msg.body.invite.door));
   GNUNET_memcpy(&(msg.body.invite.key), &key, sizeof(msg.body.invite.key));
@@ -1410,6 +1675,8 @@ GNUNET_CHAT_context_send_text (struct GNUNET_CHAT_Context *context,
     return GNUNET_SYSERR;
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
   msg.header.kind = GNUNET_MESSENGER_KIND_TEXT;
   msg.body.text.text = GNUNET_strdup(text);
 
@@ -1431,6 +1698,8 @@ GNUNET_CHAT_context_send_read_receipt (struct GNUNET_CHAT_Context *context,
 
   char zero = '\0';
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
   msg.header.kind = GNUNET_MESSENGER_KIND_TEXT;
   msg.body.text.text = &zero;
 
@@ -1482,9 +1751,11 @@ GNUNET_CHAT_context_send_file (struct GNUNET_CHAT_Context *context,
   if (GNUNET_OK != util_hash_file(path, &hash))
     return NULL;
 
-  const char *directory = handle_get_directory(context->handle);
+  char *filename = handle_create_file_path(
+    context->handle, &hash
+  );
 
-  if (!directory)
+  if (!filename)
     return NULL;
 
   struct GNUNET_CHAT_File *file = GNUNET_CONTAINER_multihashmap_get(
@@ -1492,11 +1763,6 @@ GNUNET_CHAT_context_send_file (struct GNUNET_CHAT_Context *context,
     &hash
   );
 
-  char *filename;
-  util_get_filename (
-    directory, "files", &hash, &filename
-  );
-
   if (file)
     goto file_binding;
 
@@ -1542,10 +1808,7 @@ GNUNET_CHAT_context_send_file (struct GNUNET_CHAT_Context *context,
   bo.anonymity_level = block_anonymity_level;
   bo.content_priority = block_content_priority;
   bo.replication_level = block_replication_level;
-
-  bo.expiration_time = GNUNET_TIME_absolute_add(
-    GNUNET_TIME_absolute_get(), GNUNET_TIME_relative_get_hour_()
-  );
+  bo.expiration_time = GNUNET_TIME_absolute_get_forever_();
 
   struct GNUNET_FS_FileInformation* fi = GNUNET_FS_file_information_create_from_file(
     context->handle->fs,
@@ -1580,13 +1843,22 @@ GNUNET_CHAT_context_share_file (struct GNUNET_CHAT_Context *context,
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
-  if ((!context) || (!file) || (strlen(file->name) > NAME_MAX) ||
-      (!(context->room)))
+  if ((!context) || (!file) || 
+      (!(file->name)) || (strlen(file->name) > NAME_MAX) ||
+      (!(file->uri)) || (!(context->room)))
     return GNUNET_SYSERR;
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
   msg.header.kind = GNUNET_MESSENGER_KIND_FILE;
-  GNUNET_memcpy(&(msg.body.file.key), &(file->key), sizeof(file->key));
+
+  if (file->key)
+    GNUNET_memcpy(&(msg.body.file.key), file->key,
+                  sizeof(struct GNUNET_CRYPTO_SymmetricSessionKey));
+  else
+    memset(&(msg.body.file.key), 0, sizeof(msg.body.file.key));
+
   GNUNET_memcpy(&(msg.body.file.hash), &(file->hash), sizeof(file->hash));
   GNUNET_strlcpy(msg.body.file.name, file->name, NAME_MAX);
   msg.body.file.uri = GNUNET_FS_uri_to_string(file->uri);
@@ -1611,6 +1883,8 @@ GNUNET_CHAT_context_send_tag (struct GNUNET_CHAT_Context *context,
   char *tag_value = GNUNET_strdup(tag);
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
   msg.header.kind = GNUNET_MESSENGER_KIND_TAG;
   GNUNET_memcpy(&(msg.body.tag.hash), &(message->hash),
     sizeof(struct GNUNET_HashCode));
@@ -1627,6 +1901,54 @@ GNUNET_CHAT_context_send_tag (struct GNUNET_CHAT_Context *context,
 }
 
 
+struct GNUNET_CHAT_Discourse*
+GNUNET_CHAT_context_open_discourse (struct GNUNET_CHAT_Context *context,
+                                    const struct GNUNET_ShortHashCode *id)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!context) || (!(context->discourses)) || (!(context->room)) || (!id))
+    return NULL;
+
+  struct GNUNET_CHAT_Discourse *discourse = GNUNET_CONTAINER_multishortmap_get(
+    context->discourses, id
+  );
+
+  if (!discourse)
+  {
+    discourse = discourse_create(context, id);
+
+    if (GNUNET_OK != GNUNET_CONTAINER_multishortmap_put(context->discourses,
+        id, discourse, GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_FAST))
+    {
+      discourse_destroy(discourse);
+      return NULL;
+    }
+  }
+
+  struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
+  msg.header.kind = GNUNET_MESSENGER_KIND_SUBSCRIBE;
+  GNUNET_memcpy(&(msg.body.subscribe.discourse), id, sizeof(struct GNUNET_ShortHashCode));
+
+  const struct GNUNET_TIME_Relative subscribtion_time = GNUNET_TIME_relative_multiply(
+    GNUNET_TIME_relative_get_second_(), 10
+  );
+
+  msg.body.subscribe.time = GNUNET_TIME_relative_hton(subscribtion_time);
+  msg.body.subscribe.flags = GNUNET_MESSENGER_FLAG_SUBSCRIPTION_KEEP_ALIVE;
+
+  GNUNET_MESSENGER_send_message(
+    context->room,
+    &msg,
+    NULL
+  );
+
+  return discourse;
+}
+
+
 int
 GNUNET_CHAT_context_iterate_messages (struct GNUNET_CHAT_Context *context,
                                       GNUNET_CHAT_ContextMessageCallback callback,
@@ -1687,8 +2009,18 @@ GNUNET_CHAT_message_get_kind (const struct GNUNET_CHAT_Message *message)
       return GNUNET_CHAT_KIND_LOGIN;
     case GNUNET_CHAT_FLAG_LOGOUT:
       return GNUNET_CHAT_KIND_LOGOUT;
-    case GNUNET_CHAT_FLAG_UPDATE:
-      return GNUNET_CHAT_KIND_UPDATE;
+    case GNUNET_CHAT_FLAG_CREATE_ACCOUNT:
+      return GNUNET_CHAT_KIND_CREATED_ACCOUNT;
+    case GNUNET_CHAT_FLAG_DELETE_ACCOUNT:
+      return GNUNET_CHAT_KIND_DELETED_ACCOUNT;
+    case GNUNET_CHAT_FLAG_UPDATE_ACCOUNT:
+      return GNUNET_CHAT_KIND_UPDATE_ACCOUNT;
+    case GNUNET_CHAT_FLAG_UPDATE_CONTEXT:
+      return GNUNET_CHAT_KIND_UPDATE_CONTEXT;
+    case GNUNET_CHAT_FLAG_ATTRIBUTES:
+      return GNUNET_CHAT_KIND_ATTRIBUTES;
+    case GNUNET_CHAT_FLAG_SHARE_ATTRIBUTES:
+      return GNUNET_CHAT_KIND_SHARED_ATTRIBUTES;
     default:
       break;
   }
@@ -1878,11 +2210,16 @@ GNUNET_CHAT_message_get_text (const struct GNUNET_CHAT_Message *message)
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
-  if ((!message) || (GNUNET_YES != message_has_msg(message)))
+  if (!message)
     return NULL;
 
   if (GNUNET_CHAT_FLAG_WARNING == message->flag)
     return message->warning;
+  else if (GNUNET_CHAT_FLAG_ATTRIBUTES == message->flag)
+    return message->attr;
+
+  if (GNUNET_YES != message_has_msg(message))
+    return NULL;
 
   if (GNUNET_MESSENGER_KIND_TEXT == message->msg->header.kind)
     return message->msg->body.text.text;
@@ -1895,6 +2232,21 @@ GNUNET_CHAT_message_get_text (const struct GNUNET_CHAT_Message *message)
 }
 
 
+const struct GNUNET_CHAT_Account*
+GNUNET_CHAT_message_get_account (const struct GNUNET_CHAT_Message *message)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if (!message)
+    return NULL;
+
+  if ((message->context) && (message->context->handle))
+    return message->context->handle->current;
+  else
+    return message->account;
+}
+
+
 struct GNUNET_CHAT_File*
 GNUNET_CHAT_message_get_file (const struct GNUNET_CHAT_Message *message)
 {
@@ -1933,6 +2285,32 @@ GNUNET_CHAT_message_get_invitation (const struct GNUNET_CHAT_Message *message)
 }
 
 
+struct GNUNET_CHAT_Discourse*
+GNUNET_CHAT_message_get_discourse (const struct GNUNET_CHAT_Message *message)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!message) || (GNUNET_YES != message_has_msg(message)) ||
+      (!(message->context)))
+    return NULL;
+  
+  struct GNUNET_CHAT_Discourse *discourse;
+  
+  if (GNUNET_MESSENGER_KIND_SUBSCRIBE == message->msg->header.kind)
+    discourse = GNUNET_CONTAINER_multishortmap_get(
+      message->context->discourses,
+      &(message->msg->body.subscribe.discourse));
+  else if (GNUNET_MESSENGER_KIND_TALK == message->msg->header.kind)
+    discourse = GNUNET_CONTAINER_multishortmap_get(
+      message->context->discourses,
+      &(message->msg->body.talk.discourse));
+  else
+    discourse = NULL;
+
+  return discourse;
+}
+
+
 const struct GNUNET_CHAT_Message*
 GNUNET_CHAT_message_get_target (const struct GNUNET_CHAT_Message *message)
 {
@@ -1992,6 +2370,49 @@ GNUNET_CHAT_message_iterate_tags (const struct GNUNET_CHAT_Message *message,
 }
 
 
+uint64_t
+GNUNET_CHAT_message_available (const struct GNUNET_CHAT_Message *message)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!message) || (GNUNET_YES != message_has_msg(message)))
+    return 0;
+
+  if (GNUNET_MESSENGER_KIND_TALK == message->msg->header.kind)
+    return message->msg->body.talk.length;
+  else
+    return 0;
+}
+
+
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_message_read (const struct GNUNET_CHAT_Message *message,
+                          char *data,
+                          uint64_t size)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!message) || (GNUNET_YES != message_has_msg(message)))
+    return GNUNET_SYSERR;
+
+  if (GNUNET_MESSENGER_KIND_TALK != message->msg->header.kind)
+    return GNUNET_SYSERR;
+
+  const uint64_t available = message->msg->body.talk.length;
+
+  if (available < size)
+    return GNUNET_NO;
+
+  GNUNET_memcpy(
+    data,
+    message->msg->body.talk.data,
+    size
+  );
+
+  return GNUNET_OK;
+}
+
+
 const char*
 GNUNET_CHAT_file_get_name (const struct GNUNET_CHAT_File *file)
 {
@@ -2036,16 +2457,13 @@ GNUNET_CHAT_file_get_local_size (const struct GNUNET_CHAT_File *file)
   if (!file)
     return 0;
 
-  const char *directory = handle_get_directory(file->handle);
-
-  if (!directory)
-    return 0;
-
-  char *filename;
-  util_get_filename (
-    directory, "files", &(file->hash), &filename
+  char *filename = handle_create_file_path(
+    file->handle, &(file->hash)
   );
 
+  if (!filename)
+    return 0;
+
   uint64_t size;
   if (GNUNET_OK != GNUNET_DISK_file_size(filename, &size, GNUNET_NO, GNUNET_YES))
     size = 0;
@@ -2055,6 +2473,18 @@ GNUNET_CHAT_file_get_local_size (const struct GNUNET_CHAT_File *file)
 }
 
 
+struct GNUNET_CHAT_Uri*
+GNUNET_CHAT_file_get_uri (const struct GNUNET_CHAT_File *file)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!file) || (!(file->uri)))
+    return NULL;
+
+  return uri_create_file(file->uri);
+}
+
+
 enum GNUNET_GenericReturnValue
 GNUNET_CHAT_file_is_uploading (const struct GNUNET_CHAT_File *file)
 {
@@ -2067,6 +2497,24 @@ GNUNET_CHAT_file_is_uploading (const struct GNUNET_CHAT_File *file)
 }
 
 
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_file_is_ready (const struct GNUNET_CHAT_File *file)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!file) || (file->status & GNUNET_CHAT_FILE_STATUS_MASK))
+    return GNUNET_NO;
+
+  const uint64_t size = GNUNET_CHAT_file_get_size(file);
+  const uint64_t local_size = GNUNET_CHAT_file_get_local_size(file);
+
+  if (size != local_size)
+    return GNUNET_NO;
+  else
+    return GNUNET_YES;
+}
+
+
 const char*
 GNUNET_CHAT_file_open_preview (struct GNUNET_CHAT_File *file)
 {
@@ -2078,20 +2526,25 @@ GNUNET_CHAT_file_open_preview (struct GNUNET_CHAT_File *file)
   if (file->preview)
     return file->preview;
 
-  const char *directory = handle_get_directory(file->handle);
-
-  if (!directory)
-    return NULL;
-
-  char *filename;
-  util_get_filename (
-    directory, "files", &(file->hash), &filename
+  char *filename = handle_create_file_path(
+    file->handle, &(file->hash)
   );
 
+  if (!filename)
+    return NULL;
+
   if (GNUNET_YES != GNUNET_DISK_file_test(filename))
     goto free_filename;
 
-  file->preview = GNUNET_DISK_mktemp(file->name);
+  if (!(file->key))
+  {
+    file->preview = filename;
+    return file->preview;
+  }
+
+  file->preview = GNUNET_DISK_mktemp(
+    file->name? file->name : ""
+  );
 
   if (!(file->preview))
     goto free_filename;
@@ -2100,7 +2553,7 @@ GNUNET_CHAT_file_open_preview (struct GNUNET_CHAT_File *file)
 
   if ((GNUNET_OK != GNUNET_DISK_file_copy(filename, file->preview)) ||
       (GNUNET_OK != util_decrypt_file(file->preview,
-				      &(file->hash), &(file->key))))
+      &(file->hash), file->key)))
   {
     GNUNET_free(file->preview);
     file->preview = NULL;
@@ -2120,8 +2573,22 @@ GNUNET_CHAT_file_close_preview (struct GNUNET_CHAT_File *file)
   if ((!file) || (!(file->preview)))
     return;
 
-  remove(file->preview);
+  if (!(file->key))
+    goto skip_filename;
 
+  char *filename = handle_create_file_path(
+    file->handle, &(file->hash)
+  );
+
+  if (!filename)
+    goto skip_filename;
+
+  if (0 != strcmp(filename, file->preview))
+    remove(file->preview);
+
+  GNUNET_free(filename);
+
+skip_filename:
   GNUNET_free(file->preview);
   file->preview = NULL;
 }
@@ -2182,18 +2649,15 @@ GNUNET_CHAT_file_start_download (struct GNUNET_CHAT_File *file,
     return GNUNET_OK;
   }
 
-  const char *directory = handle_get_directory(file->handle);
+  char *filename = handle_create_file_path(
+    file->handle, &(file->hash)
+  );
 
-  if (!directory)
+  if (!filename)
     return GNUNET_SYSERR;
 
   const uint64_t size = GNUNET_FS_uri_chk_get_file_size(file->uri);
 
-  char *filename;
-  util_get_filename (
-    directory, "files", &(file->hash), &filename
-  );
-
   uint64_t offset;
   if (GNUNET_OK != GNUNET_DISK_file_size(filename, &offset, 
       GNUNET_NO, GNUNET_YES))
@@ -2204,7 +2668,7 @@ GNUNET_CHAT_file_start_download (struct GNUNET_CHAT_File *file,
     if (callback)
       callback(cls, file, size, size);
 
-    return GNUNET_OK;
+    goto free_filename;
   }
 
   file_bind_downlaod(file, callback, cls);
@@ -2228,6 +2692,7 @@ GNUNET_CHAT_file_start_download (struct GNUNET_CHAT_File *file,
   if (file->download)
     file->status |= GNUNET_CHAT_FILE_STATUS_DOWNLOAD;
 
+free_filename:
   GNUNET_free(filename);
   return GNUNET_OK;
 }
@@ -2307,16 +2772,13 @@ GNUNET_CHAT_file_unindex (struct GNUNET_CHAT_File *file,
   if (file->unindex)
     return GNUNET_OK;
 
-  const char *directory = handle_get_directory(file->handle);
-
-  if (!directory)
-    return GNUNET_SYSERR;
-
-  char *filename;
-  util_get_filename (
-    directory, "files", &(file->hash), &filename
+  char *filename = handle_create_file_path(
+    file->handle, &(file->hash)
   );
 
+  if (!filename)
+    return GNUNET_SYSERR;
+
   file->unindex = GNUNET_FS_unindex_start(
     file->handle->fs, filename, file
   );
@@ -2365,6 +2827,7 @@ GNUNET_CHAT_invitation_reject (struct GNUNET_CHAT_Invitation *invitation)
     return;
 
   struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
 
   msg.header.kind = GNUNET_MESSENGER_KIND_TAG;
   GNUNET_memcpy(&(msg.body.tag.hash), &(invitation->hash),
@@ -2411,32 +2874,142 @@ GNUNET_CHAT_invitation_is_rejected (const struct GNUNET_CHAT_Invitation *invitat
 }
 
 
-const struct GNUNET_CHAT_Contact*
-GNUNET_CHAT_ticket_get_contact (const struct GNUNET_CHAT_Ticket *ticket)
+const struct GNUNET_ShortHashCode*
+GNUNET_CHAT_discourse_get_id (const struct GNUNET_CHAT_Discourse *discourse)
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
-  if (!ticket)
+  if (!discourse)
     return NULL;
 
-  struct GNUNET_ShortHashCode shorthash;
-  util_shorthash_from_member(ticket->issuer, &shorthash);
+  return &(discourse->id);
+}
 
-  return GNUNET_CONTAINER_multishortmap_get(
-    ticket->handle->contacts, &shorthash
-  );
+
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_discourse_is_open (const struct GNUNET_CHAT_Discourse *discourse)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if (!discourse)
+    return GNUNET_SYSERR;
+
+  struct GNUNET_CHAT_DiscourseSubscription *sub;
+  for (sub = discourse->head; sub; sub = sub->next)
+  {
+    if (GNUNET_TIME_absolute_cmp(sub->end, <, GNUNET_TIME_absolute_get()))
+      continue;
+
+    if (GNUNET_YES == sub->contact->owned)
+      return GNUNET_YES;
+  }
+
+  return GNUNET_NO;
 }
 
 
 void
-GNUNET_CHAT_ticket_consume (struct GNUNET_CHAT_Ticket *ticket,
-                            GNUNET_CHAT_TicketAttributeCallback callback,
-                            void *cls)
+GNUNET_CHAT_discourse_close (struct GNUNET_CHAT_Discourse *discourse)
 {
   GNUNET_CHAT_VERSION_ASSERT();
 
-  if (!ticket)
+  if ((!discourse) || (!(discourse->context)) || (!(discourse->context->room)))
     return;
+  
+  struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
 
-  ticket_consume(ticket, callback, cls);
+  msg.header.kind = GNUNET_MESSENGER_KIND_SUBSCRIBE;
+
+  GNUNET_memcpy(
+    &(msg.body.subscribe.discourse),
+    &(discourse->id),
+    sizeof(struct GNUNET_ShortHashCode)
+  );
+
+  msg.body.subscribe.time = GNUNET_TIME_relative_hton(GNUNET_TIME_relative_get_zero_());
+  msg.body.subscribe.flags = GNUNET_MESSENGER_FLAG_SUBSCRIPTION_UNSUBSCRIBE;
+
+  GNUNET_MESSENGER_send_message(
+    discourse->context->room,
+    &msg,
+    NULL
+  );
+}
+
+
+enum GNUNET_GenericReturnValue
+GNUNET_CHAT_discourse_write (struct GNUNET_CHAT_Discourse *discourse,
+                             const char *data,
+                             uint64_t size)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if ((!discourse) || (!data) || (!(discourse->context)) || 
+      (!(discourse->context->room)))
+    return GNUNET_SYSERR;
+  
+  static const uint64_t max_size = (uint16_t) (
+    GNUNET_MAX_MESSAGE_SIZE - GNUNET_MIN_MESSAGE_SIZE -
+    sizeof (struct GNUNET_MESSENGER_Message)
+  );
+
+  struct GNUNET_MESSENGER_Message msg;
+  memset(&msg, 0, sizeof(msg));
+
+  msg.header.kind = GNUNET_MESSENGER_KIND_TALK;
+  msg.body.talk.data = GNUNET_malloc(size > max_size? max_size : size);
+
+  GNUNET_memcpy(
+    &(msg.body.talk.discourse),
+    &(discourse->id),
+    sizeof(struct GNUNET_ShortHashCode)
+  );
+
+  while (size > 0)
+  {
+    msg.body.talk.length = (uint16_t) (size > max_size? max_size : size);
+
+    GNUNET_memcpy(
+      msg.body.talk.data,
+      data,
+      msg.body.talk.length
+    );
+
+    size -= msg.body.talk.length;
+    data += msg.body.talk.length;
+
+    GNUNET_MESSENGER_send_message(discourse->context->room, &msg, NULL);
+  }
+
+  GNUNET_free(msg.body.talk.data);
+  return GNUNET_OK;
+}
+
+
+int
+GNUNET_CHAT_discourse_iterate_contacts (const struct GNUNET_CHAT_Discourse *discourse,
+                                        GNUNET_CHAT_DiscourseContactCallback callback,
+                                        void *cls)
+{
+  GNUNET_CHAT_VERSION_ASSERT();
+
+  if (! discourse)
+    return GNUNET_SYSERR;
+
+  int result = 0;
+
+  struct GNUNET_CHAT_DiscourseSubscription *sub;
+  for (sub = discourse->head; sub; sub = sub->next)
+  {
+    if (GNUNET_TIME_absolute_cmp(sub->end, <, GNUNET_TIME_absolute_get()))
+      continue;
+
+    if (callback)
+      callback(cls, discourse, sub->contact);
+
+    result++;
+  }
+
+  return result;
 }
