@@ -34,16 +34,7 @@ cont_lobby_write_records (void *cls,
 
   lobby->query = NULL;
 
-  const struct GNUNET_HashCode *key = GNUNET_MESSENGER_room_get_key(
-    lobby->context->room
-  );
-
-  char *name;
-  util_lobby_name(key, &name);
-
-  handle_delete_account(lobby->handle, name);
-
-  GNUNET_free(name);
+  handle_delete_lobby(lobby->handle, lobby);
 
   if (GNUNET_EC_NONE == ec)
   {
@@ -53,6 +44,7 @@ cont_lobby_write_records (void *cls,
 
   handle_send_internal_message(
     lobby->handle,
+    NULL,
     lobby->context,
     GNUNET_CHAT_FLAG_WARNING,
     GNUNET_ErrorCode_get_hint(ec)
@@ -83,6 +75,7 @@ cont_lobby_identity_create (void *cls,
   {
     handle_send_internal_message(
 	    lobby->handle,
+      NULL,
     	lobby->context,
     	GNUNET_CHAT_FLAG_WARNING,
 	    GNUNET_ErrorCode_get_hint(ec)
@@ -115,13 +108,13 @@ cont_lobby_identity_create (void *cls,
   char *label;
   util_get_context_label(lobby->context->type, key, &label);
 
-  lobby->uri = uri_create(&public_zone, label);
+  lobby->uri = uri_create_chat(&public_zone, label);
   GNUNET_free(label);
 
   lobby->query = GNUNET_NAMESTORE_record_set_store(
     lobby->handle->namestore,
     zone,
-    lobby->uri->label,
+    lobby->uri->chat.label,
     1,
     data,
     cont_lobby_write_records,
