@@ -47,17 +47,25 @@ enum GNUNET_CHAT_MessageFlag
   GNUNET_CHAT_FLAG_REFRESH = 2,
   GNUNET_CHAT_FLAG_LOGIN = 3,
   GNUNET_CHAT_FLAG_LOGOUT = 4,
-  GNUNET_CHAT_FLAG_UPDATE = 5
+  GNUNET_CHAT_FLAG_CREATE_ACCOUNT = 5,
+  GNUNET_CHAT_FLAG_DELETE_ACCOUNT = 6,
+  GNUNET_CHAT_FLAG_UPDATE_ACCOUNT = 7,
+  GNUNET_CHAT_FLAG_UPDATE_CONTEXT = 8,
+  GNUNET_CHAT_FLAG_ATTRIBUTES = 9,
+  GNUNET_CHAT_FLAG_SHARE_ATTRIBUTES = 10
 };
 
 struct GNUNET_CHAT_Message
 {
+  const struct GNUNET_CHAT_Account *account;
+
   struct GNUNET_CHAT_Context *context;
   struct GNUNET_SCHEDULER_Task *task;
 
   union {
     const struct GNUNET_MESSENGER_Message *msg;
     const char *warning;
+    const char *attr;
   };
 
   struct GNUNET_HashCode hash;
@@ -84,16 +92,18 @@ message_create_from_msg (struct GNUNET_CHAT_Context *context,
 
 /**
  * Creates an internal chat message with an optional chat
- * <i>context</i>, a custom <i>flag</i> and an optional
- * <i>warning</i> text.
+ * <i>account</i> or <i>context</i>, a custom <i>flag</i> 
+ * and an optional <i>warning</i> text.
  *
+ * @param[in] account Chat account or NULL
  * @param[in,out] context Chat context or NULL
  * @param[in] flag Chat message flag
  * @param[in] warning Warning text
  * @return New internal chat message
  */
 struct GNUNET_CHAT_Message*
-message_create_internally (struct GNUNET_CHAT_Context *context,
+message_create_internally (const struct GNUNET_CHAT_Account *account,
+                           struct GNUNET_CHAT_Context *context,
                            enum GNUNET_CHAT_MessageFlag flag,
                            const char *warning);
 
